/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Control command for kernel function of User level translator
 *  from IPv6 to IPv4.
 * Usage: faith [-h] [prefix] [on|off]
 * Usage: faith
 *        faith -h
 *        faith fec0:0:0:1:2:3:0:0 on
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <err.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/resource.h>

#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

int	pflag;

static void
usage(void)
{

	(void)fprintf(stderr,
		      "usage:\n%s",
		      "\tfaith [prefix] [on|off]\n"
		      "\tfaith -h\n"
		      "\tfaith\n"
		      );
	exit(1);
}

int
faith_disp(void)
{
	char prefix_str[INET6_ADDRSTRLEN];
	struct in6_addr prefix = IN6ADDR_ANY_INIT;
	int keepfaith = 0, mib[4], size;

	mib[0] = CTL_NET;
	mib[1] = PF_INET6;
	mib[2] = 0;
	mib[3] = IPV6CTL_FAITH_PREFIX;

	size = sizeof(struct in6_addr);
	if (sysctl(mib, 4, &prefix, &size, NULL, 0) < 0) {
		printf("%s\n", strerror(errno));
		return (errno);
	}
	inet_ntop(AF_INET6,
		  &prefix,
		  prefix_str,
		  sizeof(prefix_str) - 1);
	printf("faith_prefix: %s\n", prefix_str);

	mib[3] = IPV6CTL_KEEPFAITH;
	size = sizeof(int);
	if (sysctl(mib, 4, &keepfaith, &size, NULL, 0) < 0) {
		printf("%s\n", strerror(errno));
		return (errno);
	}
	printf("faith: %s\n", keepfaith ? "on" : "off");

	return (errno);
}

int
main(int argc, char **argv)
{
	int ch, mib[4], help = 0, on, onoff = 0, size;
	struct in6_addr ip6_faith_prefix = IN6ADDR_ANY_INIT;

	if (argc == 1)
		exit (faith_disp());

	while ((ch = getopt(argc, argv, "h")) !=  -1) {
		switch (ch) {
		case 'h': help = 1; break;
		default:
			usage();
			/*NOTREACHED*/
		}
	}
	argc -= optind;
	argv += optind;

	/* option process */
	if (help == 1) {
		usage();
		/*NOTREACHED*/
	}
	/* optional arg process */
	if (argc == 0)
		exit (0);
	if (argc && inet_pton(AF_INET6, *argv, &ip6_faith_prefix) == 1) {
		mib[0] = CTL_NET;
		mib[1] = PF_INET6;
		mib[2] = 0;
		mib[3] = IPV6CTL_FAITH_PREFIX;
		size = sizeof(struct in6_addr);
		if (sysctl(mib, 4, 0, 0, &ip6_faith_prefix, size) < 0) {
			printf("%s\n", strerror(errno));
			exit(errno);
		}
		onoff = 1;
		on = 1;
		argc--;
		argv++;
	}
	if (argc) {
		if (strcmp(*argv, "on") == 0 ||
		    strcmp(*argv, "1") == 0 ||
		    strcmp(*argv, "yes") == 0 ||
		    strcmp(*argv, "y") == 0)
			on = 1;
		else if (strcmp(*argv, "off") == 0 ||
			 strcmp(*argv, "0") == 0 ||
			 strcmp(*argv, "no") == 0 ||
			 strcmp(*argv, "n") == 0)
			on = 0;
		else {
			usage();
			/*NOTREACHED*/
		}
		onoff = 1;
		argc--;
		argv++;
	}
	if (onoff == 1) {
		mib[0] = CTL_NET;
		mib[1] = PF_INET6;
		mib[2] = 0;
		mib[3] = IPV6CTL_KEEPFAITH;
		size = sizeof(int);
		if (sysctl(mib, 4, 0, 0, &on, size) < 0) {
			printf("%s\n", strerror(errno));
			exit(errno);
		}
	}
	exit (faith_disp());
}
