/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "faithd.h"

char ftpbuf[MSS];

void
ftp_relay(int s_src, int s_dst)
{
	ssize_t n;
	fd_set readfds;
	int err, s_rcv = s_src, s_snd = s_dst, s_dat, s_dat6;
	int ctl = YES, data_exist = NO;
	int atmark;

	syslog(LOG_INFO, "starting ftp control connection");

	FD_ZERO(&readfds);

	for (;;) {
		FD_SET(s_src, &readfds);
		FD_SET(s_dst, &readfds);
		if (data_exist == YES) {
			FD_SET(s_dat, &readfds);
			FD_SET(s_dat6, &readfds);
		}

		err = select(256, &readfds, NULL, NULL, NULL);
		if (err == -1)
			exit_failure("select: %s", ERRSTR);

		if (FD_ISSET(s_src, &readfds)) {
			ctl = YES;
			s_rcv = s_src;
			s_snd = s_dst;
		}
		if (FD_ISSET(s_dst, &readfds)) {
			ctl = YES;
			s_rcv = s_dst;
			s_snd = s_src;
		}
		if (data_exist == YES && FD_ISSET(s_dat, &readfds)) {
			ctl = NO;
			s_rcv = s_dat;
			s_snd = s_dat6;
		}
		if (data_exist == YES && FD_ISSET(s_dat6, &readfds)) {
			ctl = NO;
			s_rcv = s_dat6;
			s_snd = s_dat;
		}

		err = ioctl(s_rcv, SIOCATMARK, &atmark);
		if (err != -1 && atmark == 1) {
			n = read(s_rcv, ftpbuf, 1);
			if (n == -1) goto bad;
			send(s_snd, ftpbuf, n, MSG_OOB);
			n = read(s_rcv, ftpbuf, sizeof(ftpbuf));
			if (n == -1) goto bad;
			write(s_snd, ftpbuf, n);
			continue;
		}

		n = read(s_rcv, ftpbuf, sizeof(ftpbuf));

		if (ctl == YES && s_rcv == s_src) {
			if (memcmp("LPRT", ftpbuf, 4) == 0) {
				err =ftp_active(s_src, s_dst, &s_dat, &s_dat6);
				if (err == RETURN_SUCCESS)
					data_exist = YES;
				else
					data_exist = NO;
				continue;
			}
			if (memcmp("LPSV", ftpbuf, 4) == 0) {
				err=ftp_passive(s_src, s_dst, &s_dat, &s_dat6);
				if (err == RETURN_SUCCESS)
					data_exist = YES;
				else
					data_exist = NO;
				continue;
			}
		}
		switch (n) {
		case -1:
			goto bad;
		case 0:
			close(s_rcv);
			close(s_snd);
			FD_CLR(s_rcv, &readfds);
			FD_CLR(s_snd, &readfds);
			if (ctl == YES)
				exit_success("terminating ftp control connection");
			else {
				data_exist = NO;
				syslog(LOG_INFO, "terminating ftp data connection");
			}
			break;
		default:
			write(s_snd, ftpbuf, n);
		}
	}
	/* never reached */

 bad:
	exit_failure(ERRSTR);
}

int
ftp_active(int s_src, int s_dst, int *s_datp, int *s_dat6p)
{
	ssize_t n;
	fd_set readfds;
	int scan, len, s_wld, err;
	int s_dat, s_dat6;
	unsigned int af;
	unsigned int hal, h1, h2, h3, h4, h5, h6, h7, h8;
	unsigned int h9, h10, h11, h12, h13, h14, h15, h16;
	unsigned int pal, p1,p2;
	struct sockaddr_in6 dataddr6;
	struct sockaddr_in wldaddr, dataddr;
	char *a, *p;

	/* save LPRT */
	scan = sscanf(ftpbuf + 5,
		      "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
		      &af, &hal, &h1, &h2, &h3, &h4,
		      &h5, &h6, &h7, &h8,
		      &h9, &h10, &h11, &h12,
		      &h13, &h14, &h15, &h16,
		      &pal, &p1, &p2);
	if (scan != 21) goto bad1;
	memset(&dataddr6, 0, sizeof(dataddr6));
	dataddr6.sin6_len = sizeof(dataddr6);
	dataddr6.sin6_family = AF_INET6;
	dataddr6.sin6_addr.s6_addr32[0] = htonl((h1<<24)|(h2<<16)|(h3<<8)|h4);
	dataddr6.sin6_addr.s6_addr32[1] = htonl((h5<<24)|(h6<<16)|(h7<<8)|h8);
	dataddr6.sin6_addr.s6_addr32[2] = 
		htonl((h9<<24)|(h10<<16)|(h11<<8)|h12);
	dataddr6.sin6_addr.s6_addr32[3] = 
		htonl((h13<<24)|(h14<<16)|(h15<<8)|h16);
	dataddr6.sin6_port = htons((p1<<8)|p2);

	/* LPRT -> PORT */
	len = sizeof(wldaddr);
	err = getsockname(s_dst, (struct sockaddr *)&wldaddr, &len);
	if (err == -1) goto bad1;
	wldaddr.sin_port = 0;
	s_wld = socket(AF_INET, SOCK_STREAM, 0);
	if (s_wld == -1) goto bad1;
	err = bind(s_wld, (struct sockaddr *)&wldaddr, len);
	if (err == -1) goto bad2;
	err = listen(s_wld, 1);
	if (err == -1) goto bad2;

	err = getsockname(s_wld, (struct sockaddr *)&wldaddr, &len);
	if (err == -1) goto bad2;
	a = (char *)&wldaddr.sin_addr;
	p = (char *)&wldaddr.sin_port;
	n = sprintf(ftpbuf, "PORT %d,%d,%d,%d,%d,%d\r\n",
			  UC(a[0]),UC(a[1]),UC(a[2]),UC(a[3]),
			  UC(p[0]), UC(p[1]));
	write(s_dst, ftpbuf, n);
		
	/* PORT OK */
	FD_ZERO(&readfds);
	FD_SET(s_dst, &readfds);
	err = select(256, &readfds, NULL, NULL, NULL);
	if (err == -1) goto bad2;
	n = read(s_dst, ftpbuf, sizeof(ftpbuf));
	if (memcmp("200 PORT", ftpbuf, 8) != 0) {
		write(s_src, ftpbuf, n);
		goto bad2;
	}
	ftpbuf[4] = 'L';
	ftpbuf[5] = 'P';
	write(s_src, ftpbuf, n);

	/* e.g. LIST */
	FD_ZERO(&readfds);
	FD_SET(s_src, &readfds);
	err = select(256, &readfds, NULL, NULL, NULL);
	if (err == -1) goto bad2;
	n = read(s_src, ftpbuf, sizeof(ftpbuf));
	write(s_dst, ftpbuf, n);
		
	/* Now make a data connection */
	/* from the server */
	len = sizeof(dataddr);
	s_dat = accept(s_wld, (struct sockaddr *)&dataddr, &len);
	if (s_dat == -1) goto bad2;
	/* to the client */
	s_dat6 = socket(AF_INET6, SOCK_STREAM, 0);
	if (s_dat6 == -1) goto bad3;
#if 0
	err = setsockopt(s_wld, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if (err == -1) goto bad3;
	myaddr6.sin6_port = htons(20);
#endif
	err = connect(s_dat6, (struct sockaddr *)&dataddr6, sizeof(dataddr6));
	if (err == -1) goto bad4;
	close(s_wld);
	
	*s_datp = s_dat;
	*s_dat6p = s_dat6;
	syslog(LOG_INFO, "starting ftp data connection");
	return RETURN_SUCCESS;

 bad4:
	close(s_dat6);
 bad3:
	close(s_dat);
 bad2:
	close(s_wld);
 bad1:
	/* probably, we should return an error message */
	return RETURN_FAILURE;
}

int
ftp_passive(int s_src, int s_dst, int *s_datp, int *s_dat6p)
{
	ssize_t n;
	fd_set readfds;
	int scan, len, s_wld6, err;
	int s_dat, s_dat6;
	unsigned int h1, h2, h3, h4;
	unsigned int p1,p2;
	struct sockaddr_in6 wldaddr6, dataddr6, srcaddr6;
	struct sockaddr_in dataddr;
	char *a, *p, *c;

	/* LPSV -> PASV */
	n = sprintf(ftpbuf, "PASV\r\n");
	write(s_dst, ftpbuf, n);

	/* 227 PASV (a,b,c,d,e,f)*/
	FD_ZERO(&readfds);
	FD_SET(s_dst, &readfds);
	err = select(256, &readfds, NULL, NULL, NULL);
	if (err == -1) goto bad1;
	n = read(s_dst, ftpbuf, sizeof(ftpbuf));
	ftpbuf[n] = '\0'; /* xxx what if n == MSS? */
	if (memcmp(ftpbuf, "227", 3) != 0) goto bad1;
	c = strchr(ftpbuf, '(');
	if (c == NULL) goto bad1;
	scan = sscanf(c,"(%d,%d,%d,%d,%d,%d)", &h1, &h2, &h3, &h4, &p1, &p2);
	if (scan != 6) goto bad1;
	memset(&dataddr, 0, sizeof(dataddr));
	dataddr.sin_len = sizeof(dataddr);
	dataddr.sin_family = AF_INET;	
	dataddr.sin_addr.s_addr = htonl((h1<<24)|(h2<<16)|(h3<<8)|h4);
	dataddr.sin_port = htons((p1<<8)|p2);

	/* LPSV */
	s_wld6 = socket(AF_INET6, SOCK_STREAM, 0);
	if (s_wld6 == -1) goto bad2;
	len = sizeof(wldaddr6);
	memset(&wldaddr6, 0, sizeof(wldaddr6));
	wldaddr6.sin6_len = len;
	wldaddr6.sin6_family = AF_INET6;
	/*
	 * We never call getsockname because s_src is a fake connection.
	 * Its local address is the IPv4 compatible address which is not
	 * mine. So we can't bind. Rather, we create a nomarl wildcard
	 * socket!
	 */
	err = bind(s_wld6, (struct sockaddr *)&wldaddr6, len);
	if (err == -1) goto bad3;
	err = listen(s_wld6, 1);
	if (err == -1) goto bad3;
	/*
	 * port should be from wldaddr6.
	 * but address should be from s_src.
	 */
	err = getsockname(s_wld6, (struct sockaddr *)&wldaddr6, &len);
	if (err == -1) goto bad3;
	len = sizeof(srcaddr6);
	err = getsockname(s_src, (struct sockaddr *)&srcaddr6, &len);
	if (err == -1) goto bad3;
	a = (char *)&srcaddr6.sin6_addr;
	p = (char *)&wldaddr6.sin6_port;
	n = sprintf(ftpbuf, 
			  "228 Entering Long Passive Mode (%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)\r\n",
			  6, 16, UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
			  UC(a[4]), UC(a[5]), UC(a[6]), UC(a[7]),
			  UC(a[8]), UC(a[9]), UC(a[10]), UC(a[11]),
			  UC(a[12]), UC(a[13]), UC(a[14]), UC(a[15]),
			  2, UC(p[0]), UC(p[1]));
	write(s_src, ftpbuf, n);

	/* e.g. LIST */
	FD_ZERO(&readfds);
	FD_SET(s_src, &readfds);
	err = select(256, &readfds, NULL, NULL, NULL);
	if (err == -1) goto bad3;
	n = read(s_src, ftpbuf, sizeof(ftpbuf));
	write(s_dst, ftpbuf, n);

	/* Now make a data connection */
	/* from the client */
	len = sizeof(dataddr6);
	s_dat6 = accept(s_wld6, (struct sockaddr *)&dataddr6, &len);
	if (s_dat6 == -1)  goto bad3;
	/* to the server */
	s_dat = socket(AF_INET, SOCK_STREAM, 0);
	if (s_dat == -1) goto bad4;
	err = connect(s_dat, (struct sockaddr *)&dataddr, sizeof(dataddr));
	if (err == -1) goto bad5;

	close(s_wld6);

	*s_datp = s_dat;
	*s_dat6p = s_dat6;
	syslog(LOG_INFO, "starting ftp data connection");
	return RETURN_SUCCESS;

 bad5:
	close(s_dat);
 bad4:
	close(s_dat6);
 bad3:
	close(s_wld6);
 bad2:
	return RETURN_FAILURE;
 bad1:
	write(s_src, ftpbuf, n);
	return RETURN_FAILURE;
}

