/* 
 * $Id: parse.c,v 1.1.2.1 1998/05/07 06:15:25 sumikawa Exp $
 */

/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Copyright (c) 1997, 1998. Hitachi,Ltd.  All rights reserved. */
/* Hitachi Id: parse.c,v 1.3 1998/01/12 12:39:05 sumikawa Exp $ */

/* header include files */
#include "defs.h"
#include "parse.h"

/* local defines */
#define IN6_ADDR_ANY	"::"
#define STRING_LEN	50
#define VAR_LEN		16
#define LINE_LEN	100

/* forward references */
u_char get_key_num(char *, u_char *);
boolean check_param(char *, int *, struct rte_stat *);
void check_ctl_list(struct control **, struct in6_addr *, u_char);
void clear_ctl_list(struct control **);
struct int_config *is_if_configured(char *);
boolean check_ll_addr(struct preflist *, struct in6_addr *);

#ifdef DEBUG_PARSE
void print_config(void);
void print_ctl_list(struct control *, u_char);
#endif

struct preflist *get_if_addr(char *, struct interface **);

/* private variables declarations */
char keys[][20] = {
	"quiet", "port", "poison", "horizon", "nohorizon", "trace",
	"defaultmetric", "headerlen", "nexthop", "routetag", "auth",
	"compatible", "static", "ignore", "interface", "in", "noin",
	"out", "noout", "aggregate", "site", "nosite", "gendefault",
	"filter", "metricin", "metricout", "propagate", "srcaddr"
};

/* 
 * Func: parse_config
 * Desc: to parse the configuration file for route6d.  
 * INs : none
 * ret : none
 * Call: get_key_num()
 */
void
parse_config(void)
{
	u_char key_num;
	u_char current_state = GLOBAL;
	int current_line;

	char prefix[STRING_LEN];
	char address[STRING_LEN];

	char parameter1[VAR_LEN];
	char parameter2[VAR_LEN];
	char parameter3[VAR_LEN];	/* for STATIC */
	char key_string[VAR_LEN];
	char invalid[VAR_LEN];

	int pref_len = 0;
	int value1 = 0;
	int value2 = 0;

	struct in6_addr in6addr_any = IN6ADDR_ANY_INIT;
	struct in6_addr pref_in6;
	struct in6_addr address_in6;
	char comment = '#';

	struct static_rt *new_st_rt, *st_rt_ptr;
	struct ign_prefix *new_ign_ptr, *ign_ptr;
	struct preflist *if_ll_addr;
	struct interface *if_ptr;
	struct int_config *if_conf_ptr;
	struct aggregate *new_agr, *agr_ptr;
	struct nexthop_rte *new_nh, *nh_ptr;
	struct control *ctl_ptr;

	char line[LINE_LEN];
	char *line_end;
	boolean c_1, c_2;

	if (config_file_ptr == NULL) {
		WAIT_FOR_SIGHUP();	/* never return */
	}
	bzero((void *)line, LINE_LEN);
	current_line = 0;
	while (fgets(line, LINE_LEN, config_file_ptr) != NULL) {
		current_line++;
		bzero(parameter1, VAR_LEN);
		bzero(parameter2, VAR_LEN);
		bzero(parameter3, VAR_LEN);
		bzero(key_string, VAR_LEN);
		bzero(prefix, STRING_LEN);
		bzero(&pref_in6, sizeof(pref_in6));
		bzero(address, STRING_LEN);
		bzero(&address_in6, sizeof(address_in6));
		bzero(invalid, VAR_LEN);

		pref_len = 0;
		value1 = 0;
		value2 = 0;

		line_end = index(line, comment);
		if (line_end != NULL)
			*line_end = '\0';

		sscanf(line, "%15s", key_string); /* max 13 "defaultmetric" */
		/* VARLEN(16) = 15 + 1 ('\0') */
		if (strlen(key_string) <= 0) {
			bzero((void *)line, LINE_LEN);
			continue;
		}
		key_num = get_key_num(key_string, &current_state);
		if (key_num == INVALID_KEY)
			goto CONF_ERR;

		switch (current_state) {
		case GLOBAL:
			switch (key_num) {
			case QUIET:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;	/* never return */
				rt6_opmode = MODE_QUIET;
				difconf.int_outpass = CTL_NOSEND;
				difconf.int_ctlout = NULL;
				break;

			case PORT:	/* UDP port number */
				sscanf(line, "%15s %d %15s", key_string, &value1, invalid);
				if ((value1 <= 0) || (strlen(invalid) > 0))
					goto CONF_ERR;
				rt6_port = value1;
				break;

			case POISON:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;	/* never return */
				rt6_scheme = RT6_POISON;
				difconf.int_scheme = RT6_POISON;
				break;

			case HORIZON:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;	/* never return */
				rt6_scheme = RT6_HORZN;
				difconf.int_scheme = RT6_HORZN;
				break;

			case NOHORIZON:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;	/* never return */
				rt6_scheme = RT6_NOHORZN;
				difconf.int_scheme = RT6_NOHORZN;
				break;

			case TRACE:
				sscanf(line, "%15s %15s %15s",
				       key_string, parameter1, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;	/* never return */
				if (strcmp(parameter1, "on") == 0)
					rt6_trace = TRACE_ON;
				else if (strcmp(parameter1, "off") == 0)
					rt6_trace = TRACE_OFF;
				else
					goto CONF_ERR;
				break;

			case DEFAULTMETRIC:
				sscanf(line, "%15s %d %15s",
				       key_string, &value1, invalid);
				if ((strlen(invalid) > 0) || (value1 <= 0)
				    || (value1 > HOPCOUNT_INFINITY)) {
					goto CONF_ERR;
				}
				/* WARNING: can specify 'defaultmetric 16' ! */
				rt6_metric = value1;
				difconf.int_metric_in = value1;
				break;

			case HEADERLEN:
				sscanf(line, "%15s %d %15s",
				       key_string, &value1, invalid);
				if ((strlen(invalid) > 0) || (value1 < DEFAULT_HDRLEN)) {
					goto CONF_ERR;
				}
				rt6_hdrlen = value1;
				break;

			case NEXTHOP:
				sscanf(line, "%15s %15s %15s",
				       key_string, parameter1, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if (strcmp(parameter1, "out") == 0)
					rt6_nhopout = TRUE;
				else if (strcmp(parameter1, "noin") == 0)
					rt6_nhopnoin = TRUE;
				else
					goto CONF_ERR;
				break;

			case ROUTETAG:
				value1 = -1;
				sscanf(line, "%15s %6x %15s", key_string, &value1, invalid);
				/* permit '0x' or '0X' case */
				if ((value1 < 0) || (strlen(invalid) > 0))
					goto CONF_ERR;
				rt6_tag = value1;
				break;

			case AUTH:	/* just ignore */
				break;

			case COMPATIBLE:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				rt6_accept_compat = TRUE;
				break;
			}

			bzero((void *)line, LINE_LEN);
			break;

		case ROUTING:
			switch (key_num) {
			case STATIC:
				/* Format modified */
				/* interface token added. */
				/* static prefix gw-addr interface metric.. tagval.. */
				/* gw-addr should be linklocal */
				/* prefixlen == 0 permitted, but default is better.. */
				value1 = -1;
				value2 = -1;
				pref_len = -1;
				sscanf(line, "%15s %49s", key_string, prefix);
				if (strlen(prefix) <= 0)
					goto CONF_ERR;

				if (strcmp("default", prefix) == 0) {
					sscanf(line, "%15s %49s %49s %15s %15s %6x %15s %6x %15s",
					       key_string, prefix, address, parameter3,
					       parameter1, &value1, parameter2, &value2, invalid);
					strcpy(prefix, IN6_ADDR_ANY);
					pref_len = MAX_PREFLEN;
				} else {
					sscanf(line, "%15s %49s %d %49s %15s %15s %6x %15s %6x %15s",
					       key_string, prefix, &pref_len, address, parameter3,
					       parameter1, &value1, parameter2, &value2, invalid);
				}

				if ((strlen(invalid) > 0) ||
				    (strlen(parameter3) <= 0) ||
				    (!(IS_PTON_VALID(prefix, pref_len, pref_in6.s6_addr))) ||
				    (!(IS_VALID_ADDRESS(address, address_in6.s6_addr)))) {
					goto CONF_ERR;
				}
				/* sanity check ? */
				if (strlen(parameter1) > 0 && strlen(parameter2) > 0)
					if (strcmp(parameter1, parameter2) == 0)
						goto CONF_ERR;

				if (!IN6_IS_ADDR_LINKLOCAL(&address_in6))
					goto CONF_ERR;	/* or just ignore ? */

				if (pref_len == 0) {
					/* In fact, this is a default route */
					pref_len = MAX_PREFLEN;
					pref_in6 = in6addr_any;	
				}
				new_st_rt = (struct static_rt *)malloc(sizeof(struct static_rt));
				if (new_st_rt == NULL) {
					syslog(LOG_ERR, " %m ");
					exit_route6d();
				}
				bzero(new_st_rt, sizeof(struct static_rt));

				new_st_rt->sta_prefix.prf_addr = pref_in6;
				new_st_rt->sta_prefix.prf_len = pref_len;
				new_st_rt->sta_gw = address_in6;
				new_st_rt->sta_rtstat.rts_metric = rt6_metric;
				new_st_rt->sta_rtstat.rts_tagval = rt6_tag;
				(void)get_if_addr(parameter3, &(new_st_rt->sta_ifp));
				if (new_st_rt->sta_ifp == NULL)
					goto CONF_ERR;

				c_1 = check_param(parameter1, &value1, &(new_st_rt->sta_rtstat));
				c_2 = check_param(parameter2, &value2, &(new_st_rt->sta_rtstat));
				if (!(c_1 && c_2)) {
					free(new_st_rt);
					goto CONF_ERR;
				}
				for (st_rt_ptr = statrt; st_rt_ptr; st_rt_ptr = st_rt_ptr->sta_next) {
					if (st_rt_ptr->sta_prefix.prf_len == pref_len) {
						if (prefcmp(&(st_rt_ptr->sta_prefix.prf_addr),
							    &(new_st_rt->sta_prefix.prf_addr),
							pref_len) == TRUE) {
							/* overwrite it */
							new_st_rt->sta_next = st_rt_ptr->sta_next;
							*st_rt_ptr = *new_st_rt;
							free(new_st_rt);
							new_st_rt = NULL;
							break;
						}
					}
				}	/* for */

				if (new_st_rt != NULL) {
					new_st_rt->sta_next = statrt;
					statrt = new_st_rt;
				}
				break;

			case IGNORE:
				/* prefixlen == 0 permitted (ignore all RTE) */
				/* maybe useful in monitoring packet         */
				sscanf(line, "%15s %49s", key_string, prefix);
				if (strlen(prefix) <= 0)
					goto CONF_ERR;

				if (strcmp(prefix, "default") == 0) {
					sscanf(line, "%15s %49s %15s", key_string, prefix, invalid);
					if (strlen(invalid) > 0)
						goto CONF_ERR;
					rt6_igndefault = TRUE;
					break;	/* not added to ign-list */
				} else {
					sscanf(line, "%15s %49s %d %15s",
					       key_string, prefix, &pref_len, invalid);
					if (strlen(invalid) > 0
					    || (!IS_PTON_VALID(prefix, pref_len, pref_in6.s6_addr))) {
						goto CONF_ERR;
					}
				}

				if (pref_len == 0)	/* ignore all */
					rt6_igndefault = TRUE;
				/* also added to ign-list */
				new_ign_ptr = (struct ign_prefix *)malloc(sizeof(struct ign_prefix));
				if (new_ign_ptr == NULL) {
					syslog(LOG_ERR, " %m ");
					exit_route6d();
				}
				bzero((char *)new_ign_ptr, sizeof(struct ign_prefix));

				new_ign_ptr->igp_prefix.prf_len = pref_len;
				new_ign_ptr->igp_prefix.prf_addr = pref_in6;
				get_mask(pref_len, (char *)(new_ign_ptr->igp_mask.s6_addr));
				for (ign_ptr = ignprf; ign_ptr; ign_ptr = ign_ptr->igp_next) {
					if (ign_ptr->igp_prefix.prf_len == pref_len) {
						if (prefcmp(&(ign_ptr->igp_prefix.prf_addr),
							    &(new_ign_ptr->igp_prefix.prf_addr), pref_len)) {
							free(new_ign_ptr);
							new_ign_ptr = NULL;
							break;
						}
					}
				}
				if (new_ign_ptr != NULL) {
					new_ign_ptr->igp_next = ignprf;
					ignprf = new_ign_ptr;
				}
				break;
			}

			bzero(line, LINE_LEN);
			break;

		case INTERFACE:
			switch (key_num) {
			case INTERFACE:
				sscanf(line, "%15s %15s %15s", key_string, parameter1, invalid);
				if ((strlen(parameter1) <= 0) || (strlen(invalid) > 0))
					goto CONF_ERR;

				if_ptr = NULL;	/* forget previous interface */
				if_ll_addr = get_if_addr(parameter1, &if_ptr);

				/*
				 * maybe there is no interface but
				 * keep ifconf so that 'rescan'
				 * without conf-file change or tunnel
				 * up/down without conf-file change
				 */

				/*
				 * Even if an interface is found,
				 * if_ll_addr maybe NULL (no LinkLocal
				 * address)
				 */

				/* 
				 * If interface is already configured,
				 * overwrite it with new information,
				 * else, create new entry in link list
				 * for this interface configuration.
				 */
				if ((if_conf_ptr = is_if_configured(parameter1)) == NULL) {
					if_conf_ptr =
						(struct int_config *)malloc(sizeof(struct int_config));
					if (if_conf_ptr == NULL) {
						syslog(LOG_ERR, " %m ");
						exit_route6d();
					}
					if_conf_ptr->int_next = NULL;
					if_conf_ptr->int_scheme = rt6_scheme;
					if_conf_ptr->int_inpass = CTL_LISTEN;
					if_conf_ptr->int_ctlin = NULL;
					if_conf_ptr->int_outpass =
						(rt6_opmode == MODE_QUIET) ? CTL_NOSEND : CTL_SEND;
					if_conf_ptr->int_ctlout = NULL;

					/* 
					 * All router multicast address is
					 * default for out.
					 */
					if (rt6_opmode != MODE_QUIET) {
						ctl_ptr = (struct control *)(malloc(sizeof(struct control)));
						if (ctl_ptr == NULL) {
							free(if_conf_ptr);
							syslog(LOG_ERR, " %m ");
							exit_route6d();
						}
						bzero(ctl_ptr, sizeof(struct control));
						ctl_ptr->ctl_addr.sin6_port = htons(rt6_port);
						ctl_ptr->ctl_addr.sin6_len = sizeof(struct sockaddr_in6);
						ctl_ptr->ctl_addr.sin6_family = AF_INET6;
/* ctl_ptr->ctl_addr.sin6_flowinfo = htonl(IPV6_PRIORITY_CONTROL); */
						ctl_ptr->ctl_addr.sin6_flowinfo = 0;
						(void)inet_pton(AF_INET6, ALL_RIP6_ROUTER, address_in6.s6_addr);
						ctl_ptr->ctl_addr.sin6_addr = address_in6;
						ctl_ptr->ctl_pass = CTL_SEND;

						ctl_ptr->ctl_next = if_conf_ptr->int_ctlout;
					} else
						ctl_ptr = NULL;

					if_conf_ptr->int_ctlout = ctl_ptr;
					if_conf_ptr->int_aggr = NULL;
					if_conf_ptr->int_site = FALSE;
					if_conf_ptr->int_dstat = NULL;
					if_conf_ptr->int_dfilter = NULL;
					if_conf_ptr->int_filter = NULL;
					if_conf_ptr->int_metric_in = rt6_metric;
					if_conf_ptr->int_metric_out = DEFAULT_METRIC_OUT;
					if_conf_ptr->int_nhop = NULL;
					if_conf_ptr->int_propagate = FALSE;
					strcpy(if_conf_ptr->int_name, parameter1);

					if_conf_ptr->int_source = NULL;
					if_conf_ptr->int_next = ifconf;
					ifconf = if_conf_ptr;
				}
				break;

			case POISON:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if_conf_ptr->int_scheme = RT6_POISON;
				break;

			case HORIZON:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if_conf_ptr->int_scheme = RT6_HORZN;
				break;

			case NOHORIZON:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if_conf_ptr->int_scheme = RT6_NOHORZN;
				break;

			case IN:
				sscanf(line, "%15s %49s %15s", key_string, address, invalid);

				/* 
				 * If IN with no parameter, delete all
				 * previous entries in ctlin list, and
				 * make int_inpass to CTL_LISTEN.
				 */
				if (strlen(address) <= 0) {
					if_conf_ptr->int_inpass = CTL_LISTEN;
					clear_ctl_list(&(if_conf_ptr->int_ctlin));
					break;
				}
				if ((strlen(invalid) > 0)
				    || (!IS_VALID_ADDRESS(address, address_in6.s6_addr))
				|| (!IN6_IS_ADDR_LINKLOCAL(&address_in6))) {
					goto CONF_ERR;
				}
				/* 
				 * If address is present in the ctlin
				 * list with control CTL_LISTEN,
				 * break.  If address is present in
				 * the ctlin list with control
				 * CTL_NOLISTEN, change the control to
				 * CTL_LISTEN. If address is not
				 * present in the ctlin list, add it
				 * in * ctlin list with control
				 * CTL_LISTEN.
				 */
				check_ctl_list(&(if_conf_ptr->int_ctlin),
					       &address_in6, CTL_LISTEN);
				break;

			case NOIN:
				sscanf(line, "%15s %49s %15s",
				       key_string, address, invalid);

				/* 
				 * If NOIN with no parameter, delete
				 * all previous entries in ctlin list,
				 * and make int_inpass to CTL_NOLISTEN.
				 */
				if (strlen(address) <= 0) {
					if_conf_ptr->int_inpass = CTL_NOLISTEN;
					clear_ctl_list(&(if_conf_ptr->int_ctlin));
					break;
				}
				if ((strlen(invalid) > 0)
				    || (!IS_VALID_ADDRESS(address, address_in6.s6_addr))
				|| (!IN6_IS_ADDR_LINKLOCAL(&address_in6))) {
					goto CONF_ERR;
				}

				/* 
				 * If address is present in the ctlin
				 * list with control CTL_NOLISTEN,
				 * break. If address is present in the
				 * ctlin list with control CTL_LISTEN,
				 * change the control to CTL_NOLISTEN. 
				 * If address is not present in the
				 * ctlin list, add it in ctlin list
				 * with control CTL_NOLISTEN.
				 */
				check_ctl_list(&(if_conf_ptr->int_ctlin), &address_in6, CTL_NOLISTEN);
				break;

			case OUT:
				sscanf(line, "%15s %49s %15s", key_string, address, invalid);

				/* 
				 * If OUT with no parameter, delete
				 * all previous unicast entries in
				 * ctlout list, and make int_outpass
				 * to CTL_SEND (to FF02::9).
				 */
				if (strlen(address) <= 0) {
					if_conf_ptr->int_outpass = CTL_SEND;
					clear_ctl_list(&(if_conf_ptr->int_ctlout));
					/* rebuild FF02::9 entry */
					ctl_ptr = (struct control *)(malloc(sizeof(struct control)));
					if (ctl_ptr == NULL) {
						syslog(LOG_ERR, " %m ");
						exit_route6d();
					}
					bzero(ctl_ptr, sizeof(struct control));
					ctl_ptr->ctl_addr.sin6_port = htons(rt6_port);
					ctl_ptr->ctl_addr.sin6_len = sizeof(struct sockaddr_in6);
					ctl_ptr->ctl_addr.sin6_family = AF_INET6;
/* ctl_ptr->ctl_addr.sin6_flowinfo = htonl(IPV6_PRIORITY_CONTROL); */
					ctl_ptr->ctl_addr.sin6_flowinfo = 0;
					(void)inet_pton(AF_INET6, ALL_RIP6_ROUTER, address_in6.s6_addr);
					ctl_ptr->ctl_addr.sin6_addr = address_in6;
					ctl_ptr->ctl_pass = CTL_SEND;
					ctl_ptr->ctl_next = if_conf_ptr->int_ctlout;
					if_conf_ptr->int_ctlout = ctl_ptr;
					break;
				}
				if ((strlen(invalid) > 0) ||
				    (!IS_VALID_ADDRESS(address, address_in6.s6_addr)) ||
				    (!IN6_IS_ADDR_LINKLOCAL(&address_in6))) {
					goto CONF_ERR;
				}

				/* 
				 * If address is present in the ctlout
				 * list with control CTL_SEND, break.
				 * If address is present in the ctlout
				 * list with control CTL_NOSEND,
				 * change the control to CTL_SEND.  If
				 * address is not present in the
				 * ctlout list, add it in ctlin list
				 * with control CTL_SEND.
				 */
				check_ctl_list(&(if_conf_ptr->int_ctlout), &address_in6, CTL_SEND);
				break;

			case NOOUT:
				sscanf(line, "%15s %15s", key_string, invalid);

				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if_conf_ptr->int_outpass = CTL_NOSEND;
				clear_ctl_list(&(if_conf_ptr->int_ctlout));
				break;

			case AGGREGATE:
				/* 'aggregate default' is not allowed. */
				/* Use gendefault or ':: 0'            */
				value1 = -1;
				value2 = -1;
				pref_len = -1;
				sscanf(line, "%15s %49s %d %15s %6x %15s %6x %15s",
				  key_string, prefix, &pref_len, parameter1,
				     &value1, parameter2, &value2, invalid);

				if ((strlen(invalid) > 0) ||
				    (strlen(prefix) <= 0) ||
				    (!(IS_PTON_VALID(prefix, pref_len, pref_in6.s6_addr)))) {
					goto CONF_ERR;
				}
				if (strlen(parameter1) > 0 && strlen(parameter2) > 0)
					if (strcmp(parameter1, parameter2) == 0)
						goto CONF_ERR;

				new_agr = (struct aggregate *)(malloc(sizeof(struct aggregate)));
				if (new_agr == NULL) {
					syslog(LOG_ERR, " %m ");
					exit_route6d();
				}
				bzero((char *)new_agr, sizeof(struct aggregate));

				new_agr->agr_pref.prf_addr = pref_in6;
				new_agr->agr_pref.prf_len = pref_len;
				get_mask(pref_len, (char *)(new_agr->agr_mask.s6_addr));
				new_agr->agr_stat.rts_metric = rt6_metric;
				new_agr->agr_stat.rts_tagval = rt6_tag;
				c_1 = check_param(parameter1, &value1, &(new_agr->agr_stat));
				c_2 = check_param(parameter2, &value2, &(new_agr->agr_stat));
				if (!(c_1 && c_2)) {
					free(new_agr);
					goto CONF_ERR;
				}
				for (agr_ptr = if_conf_ptr->int_aggr; agr_ptr;
				     agr_ptr = agr_ptr->agr_next) {
					if (agr_ptr->agr_pref.prf_len >= pref_len) {
						if (prefcmp(&(agr_ptr->agr_pref.prf_addr),
							    &(new_agr->agr_pref.prf_addr), pref_len)) {
							new_agr->agr_next = agr_ptr->agr_next;
							*agr_ptr = *new_agr;
							free(new_agr);
							new_agr = NULL;
							break;
						}
					}
				}

				if (new_agr != NULL) {
					new_agr->agr_next = if_conf_ptr->int_aggr;
					if_conf_ptr->int_aggr = new_agr;
				}
				break;

			case NOSITE:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if_conf_ptr->int_site = FALSE;
				break;

			case SITE:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if_conf_ptr->int_site = TRUE;
				break;

			case GENDEFAULT:
				value1 = -1;
				value2 = -1;
				sscanf(line, " %15s %15s %6x %15s %6x %15s",
				key_string, parameter1, &value1, parameter2,
				       &value2, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;

				if (strlen(parameter1) > 0 && strlen(parameter2) > 0)
					if (strcmp(parameter1, parameter2) == 0)
						goto CONF_ERR;

				if (if_conf_ptr->int_dstat == NULL) {
					if_conf_ptr->int_dstat =
						(struct rte_stat *)(malloc(sizeof(struct rte_stat)));
					if (if_conf_ptr->int_dstat == NULL) {
						syslog(LOG_ERR, " %m ");
						exit_route6d();
					}
				}
				if_conf_ptr->int_dstat->rts_metric = rt6_metric;
				if_conf_ptr->int_dstat->rts_tagval = rt6_tag;
				c_1 = check_param(parameter1, &value1, if_conf_ptr->int_dstat);
				c_2 = check_param(parameter2, &value2, if_conf_ptr->int_dstat);
				if (!(c_1 && c_2)) {
					free(if_conf_ptr->int_dstat);
					goto CONF_ERR;
				}
				break;

			case FILTER:
				value1 = -1;
				sscanf(line, " %15s %49s ", key_string, prefix);
				if (strlen(prefix) <= 0)
					goto CONF_ERR;

				if (strcmp(prefix, "default") == 0) {
					sscanf(line, " %15s %49s %15s %6x %15s ",
					       key_string, prefix, parameter1, &value1, invalid);
					strcpy(prefix, IN6_ADDR_ANY);
					pref_len = MAX_PREFLEN;
				} else {
					sscanf(line, " %15s %49s %d %15s %6x %15s ",
					       key_string, prefix, &pref_len, parameter1, &value1, invalid);
				}

				if ((strlen(invalid) > 0)
				    || (!(IS_PTON_VALID(prefix, pref_len, pref_in6.s6_addr)))) {
					goto CONF_ERR;
				}
				if (strlen(parameter1) > 0)
					if (strcmp(parameter1, "routetag") != 0
					  || value1 < 0 || value1 > 0x0ffff)
						goto CONF_ERR;

				if ((pref_len == MAX_PREFLEN &&
				     IN6_IS_ADDR_UNSPECIFIED(&pref_in6)) ||
				    pref_len == 0) {
					/* filter all */
					if (if_conf_ptr->int_dfilter == NULL) {
						if_conf_ptr->int_dfilter =
							(struct rte_stat *)malloc(sizeof(struct rte_stat));
						if (if_conf_ptr->int_dfilter == NULL) {
							syslog(LOG_ERR, " %m ");
							exit_route6d();
						}
					}
					bzero(if_conf_ptr->int_dfilter, sizeof(struct rte_stat));
					if (strlen(parameter1) > 0) {
						if_conf_ptr->int_dfilter->rts_tagval = value1;
						if_conf_ptr->int_dfilter->rts_metric = TRUE;
					}
					if (pref_len)
						break;
					/* filter all also added to list */
					/* especially that with TAG specified */
				}
				new_agr = (struct aggregate *)(malloc(sizeof(struct aggregate)));
				if (new_agr == NULL) {
					syslog(LOG_ERR, " %m ");
					exit_route6d();
				}
				bzero(new_agr, sizeof(struct aggregate));

				new_agr->agr_pref.prf_addr = pref_in6;
				new_agr->agr_pref.prf_len = pref_len;
				get_mask(pref_len, (char *)&(new_agr->agr_mask.s6_addr));
				/* tag should be NONE if not specified */
				if (strlen(parameter1) > 0) {
					new_agr->agr_stat.rts_tagval = value1;
					new_agr->agr_stat.rts_metric = TRUE;
				}
				for (agr_ptr = if_conf_ptr->int_filter; agr_ptr;
				     agr_ptr = agr_ptr->agr_next) {
					if (agr_ptr->agr_pref.prf_len == pref_len)
						if (prefcmp(&(agr_ptr->agr_pref.prf_addr),
							    &(new_agr->agr_pref.prf_addr), pref_len))
							if (agr_ptr->agr_stat.rts_metric == FALSE
							    || agr_ptr->agr_stat.rts_tagval == value1) {
								free(new_agr);
								new_agr = NULL;
								break;
							}
				}	/* for */

				if (new_agr != NULL) {
					new_agr->agr_next = if_conf_ptr->int_filter;
					if_conf_ptr->int_filter = new_agr;
				}
				break;

			case METRICIN:
				sscanf(line, "%15s %d %15s", key_string, &value1, invalid);
				if ((strlen(invalid) > 0) || (value1 <= 0)) {
					goto CONF_ERR;
				}
				if_conf_ptr->int_metric_in = value1;
				/* metricin 1 means same as default */
				break;

			case METRICOUT:
				sscanf(line, "%15s %d %15s", key_string, &value1, invalid);
				if ((strlen(invalid) > 0) || (value1 < 0)) {
					goto CONF_ERR;
				}
				if_conf_ptr->int_metric_out = value1;
				/* metricout 0 means same as default */
				break;

			case NEXTHOP:
				sscanf(line, "%15s %49s", key_string, prefix);
				if (strcmp(prefix, "default") == 0) {
					sscanf(line, "%15s %49s %49s %15s",
					       key_string, prefix, address, invalid);
					strcpy(prefix, IN6_ADDR_ANY);
					pref_len = MAX_PREFLEN;
				} else {
					sscanf(line, "%15s %49s %d %49s %15s",
					       key_string, prefix, &pref_len, address, invalid);
				}

				if ((strlen(invalid) > 0)
				    || (!IS_PTON_VALID(prefix, pref_len, pref_in6.s6_addr))
				    || (!IS_VALID_ADDRESS(address, address_in6.s6_addr))) {
					goto CONF_ERR;
				}
				new_nh = (struct nexthop_rte *)malloc(sizeof(struct nexthop_rte));
				if (new_nh == NULL) {
					syslog(LOG_ERR, " %m ");
					exit_route6d();
				}
				bzero(new_nh, sizeof(struct nexthop_rte));

				new_nh->nh_prf.prf_addr = pref_in6;
				new_nh->nh_prf.prf_len = pref_len;
				new_nh->nh_addr = address_in6;
				get_mask(pref_len, (char *)(new_nh->nh_mask.s6_addr));

				for (nh_ptr = if_conf_ptr->int_nhop; nh_ptr; nh_ptr = nh_ptr->nh_next) {
					if (nh_ptr->nh_prf.prf_len == pref_len) {
						if (prefcmp(&(nh_ptr->nh_prf.prf_addr),
							    &(new_nh->nh_prf.prf_addr), pref_len)) {
							new_nh->nh_next = nh_ptr->nh_next;
							*nh_ptr = *new_nh;
							free(new_nh);
							new_nh = NULL;
							break;
						}
					}
				}	/* for */

				if (new_nh != NULL) {
					new_nh->nh_next = if_conf_ptr->int_nhop;
					if_conf_ptr->int_nhop = new_nh;
				}
				break;

			case PROPAGATE:
				sscanf(line, "%15s %15s", key_string, invalid);
				if (strlen(invalid) > 0)
					goto CONF_ERR;
				if_conf_ptr->int_propagate = TRUE;
				break;

			case SRCADDR:
				sscanf(line, "%15s %49s %15s", key_string, address, invalid);
				if ((strlen(invalid) > 0) ||
				    (!IS_VALID_ADDRESS(address,
						       address_in6.s6_addr)) ||
				    (!IN6_IS_ADDR_LINKLOCAL(&address_in6)) ||
				    (!check_ll_addr(if_ll_addr, &address_in6))) {
					goto CONF_ERR;
				}
				if (if_conf_ptr->int_source == NULL) {
					if_conf_ptr->int_source =
						(struct in6_addr *)malloc(sizeof(struct in6_addr));
					if (if_conf_ptr->int_source == NULL) {
						syslog(LOG_ERR, " %m ");
						exit_route6d();
					}
				}
				*(if_conf_ptr->int_source) = address_in6;
				break;
			}	/* end switch( key_num ) */

			bzero(line, LINE_LEN);
			break;
			/* because error checking is done in get_key_num() */
		}
	}

#ifdef DEBUG_PARSE
	print_config();
#endif
	return;

 CONF_ERR:
	syslog(LOG_ERR, "configuration error at line %d", current_line);
	WAIT_FOR_SIGHUP();

}

/* 
 * Func: get_key_num
 * Desc: to get key_string number for the key_string passed.
 * INs : char* key_string_name, u_char* current_state
 * ret : u_char key_number 
 * Call: none
 */
u_char
get_key_num(char *key_string_name, u_char * current_state)
{
	register int i;

	switch (*current_state) {
	case GLOBAL:
		if (strcmp(key_string_name, keys[INTERFACE]) == 0) {
			*current_state = INTERFACE;
			return (INTERFACE);
		}
		if (strcmp(key_string_name, keys[STATIC]) == 0) {
			*current_state = ROUTING;
			return (STATIC);
		}
		if (strcmp(key_string_name, keys[IGNORE]) == 0) {
			*current_state = ROUTING;
			return (IGNORE);
		}
		for (i = 0; i <= COMPATIBLE; i++) {
			if (strcmp(key_string_name, keys[i]) == 0)
				return (i);
		}
		return (INVALID_KEY);
		/* Not Reached */

	case ROUTING:
		if (strcmp(key_string_name, keys[INTERFACE]) == 0) {
			*current_state = INTERFACE;
			return (INTERFACE);
		}
		if (strcmp(key_string_name, keys[STATIC]) == 0) {
			return (STATIC);
		}
		if (strcmp(key_string_name, keys[IGNORE]) == 0) {
			return (IGNORE);
		}
		return (INVALID_KEY);
		/* Not Reached */

	case INTERFACE:
		for (i = POISON; i <= NOHORIZON; i++) {
			if (strcmp(key_string_name, keys[i]) == 0)
				return (i);
		}
		for (i = INTERFACE; i <= SRCADDR; i++) {
			if (strcmp(key_string_name, keys[i]) == 0)
				return (i);
		}
		if (strcmp(key_string_name, keys[NEXTHOP]) == 0)
			return (NEXTHOP);

		return (INVALID_KEY);
		/* Not Reached */

	}
	return (INVALID_KEY);
}

/* 
 * Func: check_param
 * Desc: to check the values passed with metric and routetag
 *       in configuration file.
 * INs : char *parameter, u_short value, struct rte_stat *rte_ptr  
 * ret : boolean 
 * Call: none 
 */
boolean
check_param(char *parameter, int *value, struct rte_stat * rte_ptr)
{
	char str[10];

	if (strlen(parameter) <= 0)
		return TRUE;

	if (strcmp(parameter, "metric") == 0) {
		if (*value <= 0)
			return FALSE;

		bzero(str, 10);
		sprintf(str, "%x", *value); /* Because value is read as %x */
		sscanf(str, "%d", value);   /* But in 'metric', value is %d */
		if (*value > HOPCOUNT_INFINITY)
			return FALSE;

		rte_ptr->rts_metric = *value;
	} else if (strcmp(parameter, "routetag") == 0) {
		if (*value < 0 || *value > 0x0ffff)
			return FALSE;
		rte_ptr->rts_tagval = *value;
	} else
		return FALSE;

	return TRUE;
}

/* 
 * Func: check_ctl_list
 * Desc: to check the address passed, is present in control list.
 * INs : struct control **list_start, char *address, u_char pass  
 * ret : none
 * Call: none
 */
void
check_ctl_list(struct control **list_start, struct in6_addr *address,
	       u_char pass)
{
	struct control *ctl_ptr;

	for (ctl_ptr = *list_start; ctl_ptr; ctl_ptr = ctl_ptr->ctl_next) {
		if (!memcmp(ctl_ptr->ctl_addr.sin6_addr.s6_addr,
			    address->s6_addr, 16)) {
			ctl_ptr->ctl_pass = pass;
			return;
		}
	}

	/* Add new entry in linklist. */
	ctl_ptr = (struct control *)(malloc(sizeof(struct control)));
	if (ctl_ptr == NULL) {
		syslog(LOG_ERR, " %m ");
		exit_route6d();
	}
	bzero(ctl_ptr, sizeof(struct control));
	ctl_ptr->ctl_addr.sin6_port = htons(rt6_port);
	ctl_ptr->ctl_addr.sin6_len = sizeof(struct sockaddr_in6);
	ctl_ptr->ctl_addr.sin6_family = AF_INET6;
	ctl_ptr->ctl_addr.sin6_flowinfo = 0;
	ctl_ptr->ctl_addr.sin6_addr = *address;
	ctl_ptr->ctl_pass = pass;
	ctl_ptr->ctl_next = *list_start;
	*list_start = ctl_ptr;
	return;
}

/* 
 * Func: clear_ctl_list
 * Desc: to clear the control list.
 * INs : struct control **list_start
 * ret : none
 * Call: none
 */
void
clear_ctl_list(struct control **list_start)
{
	struct control *ctl_ptr, *next_ptr;

	for (ctl_ptr = *list_start; ctl_ptr;) {
		next_ptr = ctl_ptr->ctl_next;
		free(ctl_ptr);
		ctl_ptr = next_ptr;
	}
	*list_start = NULL;
	return;
}

/* 
 * Func: get_if_addr
 * Desc: to get interface address for interface name passed.
 * INs : char *if_name, struct interface **if_ptr   
 * ret : struct preflist *if_ll_addr
 * Call: none
 */
struct preflist *
get_if_addr(char *if_name, struct interface **if_ptr)
{
	struct interface *tmp_if_ptr;

	for (tmp_if_ptr = ifnet; tmp_if_ptr; tmp_if_ptr = tmp_if_ptr->if_next) {
		if (strcmp(tmp_if_ptr->if_name, if_name) == 0) {
			*if_ptr = tmp_if_ptr;
			return (tmp_if_ptr->if_lladdr);
		}
	}
	return NULL;
}

/* 
 * Func: is_if_configured
 * Desc: to check interface is configured or not.
 * INs : char *if_name
 * ret : struct int_config* if_conf_ptr
 * Call: none
 */
struct int_config *
is_if_configured(char *if_name)
{
	struct int_config *if_conf_ptr;

	for (if_conf_ptr = ifconf; if_conf_ptr; if_conf_ptr = if_conf_ptr->int_next) {
		if (strcmp(if_conf_ptr->int_name, if_name) == 0)
			return (if_conf_ptr);
	}
	return NULL;
}

/* 
 * Func: check_ll_addr
 * Desc: to check interface is configured with same link-local address.
 * INs : struct preflist *if_ll_addr, char *address
 * ret : boolean
 * Call: none
 */
boolean
check_ll_addr(struct preflist * if_ll_addr, struct in6_addr * address)
{
	struct preflist *tmp_ll_addr;

	for (tmp_ll_addr = if_ll_addr; tmp_ll_addr;
	     tmp_ll_addr = tmp_ll_addr->pl_next) {
		if (!memcmp(tmp_ll_addr->pl_pref.prf_addr.s6_addr, address->s6_addr, 16))
			return TRUE;
	}
	return FALSE;
}

#ifdef DEBUG_PARSE
/* 
 * Func: print_config
 * Desc: to print all the configuration parameters.
 * INs : void
 * ret : void
 * Call: none
 */
void
print_config(void)
{
	char address1[STRING_LEN];
	char address2[STRING_LEN];

	struct static_rt *st_ptr;
	struct ign_prefix *ign_ptr;
	struct int_config *if_conf_ptr;
	struct aggregate *agr_ptr;
	struct nexthop_rte *nh_ptr;

	printf("GLOBAL parameters : \n");
	printf("------------------- \n");
	printf("rt6_opmode : %d \n", rt6_opmode);
	printf("rt6_port : %d \n", rt6_port);
	printf("rt6_scheme : %d \n", rt6_scheme);
	printf("rt6_trace : %d \n", rt6_trace);
	printf("rt6_metric : %d \n", rt6_metric);
	printf("rt6_hdrlen : %d \n", rt6_hdrlen);
	printf("rt6_nhopout : %d \n", rt6_nhopout);
	printf("rt6_nhopnoin : %d \n", rt6_nhopnoin);
	printf("rt6_tag : %x \n", rt6_tag);
	printf("rt6_accept_compat : %d \n", rt6_accept_compat);

	printf("\nROUTING parameters : \n");
	printf("-------------------- \n");

	printf("\n      static routes \n\n");
	printf(" prefix    pref_len  address  metric tag\n");
	for (st_ptr = statrt; st_ptr; st_ptr = st_ptr->sta_next) {
		bzero(address1, STRING_LEN);
		bzero(address2, STRING_LEN);
		inet_ntop(AF_INET6, st_ptr->sta_prefix.prf_addr.s6_addr,
			  address1, sizeof(address1));
		inet_ntop(AF_INET6, st_ptr->sta_gw.s6_addr, address2, sizeof(address2));
		printf(" %16s  %d  %16s  %d  %x  \n", address1, st_ptr->sta_prefix.prf_len,
		       address2, st_ptr->sta_rtstat.rts_metric,
		       st_ptr->sta_rtstat.rts_tagval);
	}
	printf("\n      ignore prefixes \n");
	printf(" prefix    pref_len \n");
	for (ign_ptr = ignprf; ign_ptr; ign_ptr = ign_ptr->igp_next) {
		bzero(address1, STRING_LEN);
		inet_ntop(AF_INET6, ign_ptr->igp_prefix.prf_addr.s6_addr,
			  address1, sizeof(address1));
		printf(" %16s  %d \n", address1, ign_ptr->igp_prefix.prf_len);
	}

	printf("\nINTERFACE parameters : \n");
	printf("-------------------- \n");

	for (if_conf_ptr = ifconf; if_conf_ptr; if_conf_ptr = if_conf_ptr->int_next) {
		printf("\n   interface %16s \n", if_conf_ptr->int_name);
		printf(" int_scheme : %d \n ", if_conf_ptr->int_scheme);
		printf(" int_inpass : %d \n ", if_conf_ptr->int_inpass);

		printf("\n      in list \n\n ");
		print_ctl_list(if_conf_ptr->int_ctlin, CTL_LISTEN);

		printf("\n      noin list \n\n ");
		print_ctl_list(if_conf_ptr->int_ctlin, CTL_NOLISTEN);

		printf("\n int_outpass : %d \n ", if_conf_ptr->int_outpass);
		printf("\n      out list \n\n ");
		print_ctl_list(if_conf_ptr->int_ctlout, CTL_SEND);

		printf("\n      aggregate list \n\n");
		printf(" prefix    pref_len  metric  tag \n");
		for (agr_ptr = if_conf_ptr->int_aggr; agr_ptr; agr_ptr = agr_ptr->agr_next) {
			bzero(address1, STRING_LEN);
			inet_ntop(AF_INET6, (agr_ptr->agr_pref.prf_addr.s6_addr),
				  address1, sizeof(address1));
			printf(" %16s  %d  %d  %x \n", address1, agr_ptr->agr_pref.prf_len,
			       agr_ptr->agr_stat.rts_metric, agr_ptr->agr_stat.rts_tagval);
		}

		printf("\n int_site : %d \n", if_conf_ptr->int_site);
		printf("\n gendefault :  \n");
		if (if_conf_ptr->int_dstat) {
			printf("\n metric : %d \n", if_conf_ptr->int_dstat->rts_metric);
			printf(" routetag : %x \n", if_conf_ptr->int_dstat->rts_tagval);
		}
		printf("\n      filter list \n\n ");
		if (if_conf_ptr->int_dfilter != NULL) {
			printf(" filter default RTE ");
			if (if_conf_ptr->int_dfilter->rts_metric == TRUE)
				printf("with routetag %x ", if_conf_ptr->int_dfilter->rts_tagval);
		}
		printf("\n prefix    pref_len  tag \n");
		for (agr_ptr = if_conf_ptr->int_filter; agr_ptr; agr_ptr = agr_ptr->agr_next) {
			bzero(address1, STRING_LEN);
			inet_ntop(AF_INET6, agr_ptr->agr_pref.prf_addr.s6_addr,
				  address1, sizeof(address1));
			printf(" %s  %d  %x \n", address1,
			       agr_ptr->agr_pref.prf_len, agr_ptr->agr_stat.rts_tagval);
		}

		printf("\n int_metric_in : %d \n", if_conf_ptr->int_metric_in);
		printf(" int_metric_out : %d \n", if_conf_ptr->int_metric_out);

		printf("\n      nexthop list \n\n");
		printf(" prefix    pref_len  address \n");
		for (nh_ptr = if_conf_ptr->int_nhop; nh_ptr; nh_ptr = nh_ptr->nh_next) {
			bzero(address1, STRING_LEN);
			inet_ntop(AF_INET6, nh_ptr->nh_prf.prf_addr.s6_addr,
				  address1, sizeof(address1));
			bzero(address2, STRING_LEN);
			inet_ntop(AF_INET6, nh_ptr->nh_addr.s6_addr, address2, sizeof(address2));
			printf(" %s  %d  %s \n", address1, nh_ptr->nh_prf.prf_len, address2);
		}

		printf("\n int_propagate : %d \n", if_conf_ptr->int_propagate);
		bzero(address1, STRING_LEN);
		if (if_conf_ptr->int_source != NULL) {
			inet_ntop(AF_INET6, if_conf_ptr->int_source->s6_addr,
				  address1, sizeof(address1));
			printf(" int_source : %s \n ", address1);
		} else
			printf(" int_source : NULL \n");
	}
	return;
}

/* 
 * Func: print_ctl_list
 * Desc: to print address in control list passed.
 * INs : struct control *list_start, u_char pass
 * ret : none
 * Call: none
 */
void
print_ctl_list(struct control *list_start, u_char pass)
{
	char address[STRING_LEN];
	struct control *ctl_ptr;

	for (ctl_ptr = list_start; ctl_ptr; ctl_ptr = ctl_ptr->ctl_next)
		if (ctl_ptr->ctl_pass == pass) {
			bzero(address, STRING_LEN);
			inet_ntop(AF_INET6, ctl_ptr->ctl_addr.sin6_addr.s6_addr,
				  address, sizeof(address));
			printf(" %s \n", address);
		}
	return;
}
#endif
