/* 
 * $Id: timer.c,v 1.1.2.1 1998/05/07 06:15:32 sumikawa Exp $
 */

/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Copyright (c) 1997, 1998. Hitachi,Ltd.  All rights reserved. */
/* Hitachi Id: timer.c,v 1.3 1998/01/12 12:39:08 sumikawa Exp $ */

/* header include files */
#include  "defs.h"

/* 
 * Func: timer()
 * Desc: To calculate time and call appropriate function.
 */
void
timer(void)
{
	int updateinterval;
	int update_time(void);

#ifdef DEBUG_MAIN
	printf("Timer Called!!!\n");
#endif				/* DEBUG_MAIN */

	if (scanning > 0) {
		scanning = 0;
#ifdef DEBUG_MAIN
		printf("Timer Aborted!!!\n");
#endif				/* DEBUG_MAIN */
		return;
	}
	(void)gettimeofday(&now, (struct timezone *)NULL);

	if (update_time())
		sendupdate = TRUE;

	if (nr_time.tv_sec <= now.tv_sec) {
		regular = 1;
	/* send_regular_update(); */

		updateinterval = TIMER_RATE + get_random_num(URANGE);
		nr_time.tv_sec = now.tv_sec + updateinterval;
		nr_time.tv_usec = now.tv_usec;
	/* next Triggered time need not be modified */
	} else {
		updateinterval = nr_time.tv_sec - now.tv_sec;
	}

#if 0
	if (sendupdate == TRUE)	/* no regular update sent */
		if (timercmp(&nt_time, &now, <))
			trigger_update();
#endif

	alarminterval = updateinterval;		/* to use in next
						 * update_time() */
	alarm(alarminterval);
	return;
}				/* end of timer() */

/* 
 * Func: get_random_num()
 * Desc: Get a random number in the specified range.
 */
int
get_random_num(int range)
{
	return ((random() % range) + 1);
}				/* end of get_random_num */

/* 
 * Func: trigger_update()
 * Desc: To maintain timer for triggered update.
 */
void
trigger_update(void)
{
	(void)gettimeofday(&now, (struct timezone *)NULL);
	send_triggered_update();

	nt_time.tv_sec = now.tv_sec + get_random_num(TRANGE);
	nt_time.tv_usec = now.tv_usec;

	return;
}				/* end of trigger_update() */

/* 
 * Func: update_time()
 * Desc: To update the timers for the route and keep their validity.
 */
int
update_time(void)
{
	struct rt_plen *rtp, *temp;
	struct gateway *gwp;
	int changed = 0;

#ifdef DEBUG_MAIN
	printf("update passed time = %d\n", alarminterval);
#endif
	for (gwp = gway; gwp; gwp = gwp->gw_next) {
		for (rtp = gwp->gw_dest; rtp;) {
			temp = rtp;
			rtp = rtp->rp_ndst;
			if ((temp->rp_state & (RTS6_KERNEL | RTS6_STATIC | RTS6_INTERFACE)) != 0)
				continue;
			temp->rp_timer += alarminterval;
#ifdef DEBUG_MAIN
			printf(" timer%d ", temp->rp_timer);
#endif
			if (temp->rp_timer > GARBAGE_TIME) {
#ifdef DEBUG_MAIN
				printf("garbage\n");
#endif
				garbage = 1;
			/* delete_local_route( temp ); */
			} else if ((temp->rp_timer > EXPIRE_TIME)
				&& (temp->rp_metric != HOPCOUNT_INFINITY)) {
#ifdef DEBUG_MAIN
				printf("expire\n");
#endif
				changed = 1;
				temp->rp_metric = HOPCOUNT_INFINITY;
				temp->rp_state |= RTS6_CHANGED;
				rt_ioctl(temp, RTM_DELETE);
			}
		}
	}

	return changed;
}				/* end of update_time() */

/* 
 * Func: timevaladd()
 * Desc: To add two timevals.
 */
void
timevaladd(struct timeval *t1, struct timeval *t2)
{
	t1->tv_sec += t2->tv_sec;
	if ((t1->tv_usec += t2->tv_usec) > 1000000) {
		t1->tv_sec++;
		t1->tv_usec -= 1000000;
	}
}				/* end of timevaladd() */

/* 
 * Func: timevalsub()
 * Desc: To subtract two timevals.
 */
void
timevalsub(struct timeval *t1, struct timeval *t2)
{
	t1->tv_sec -= t2->tv_sec;
	if ((t1->tv_usec -= t2->tv_usec) < 0) {
		t1->tv_sec--;
		t1->tv_usec += 1000000;
	}
}				/* end of timevalsub() */
