/*
 * $Id: name6.c,v 1.1.2.3.2.2 1998/05/08 09:47:46 k-sugyou Exp $
 *
 *	Atsushi Onoe <onoe@sm.sony.co.jp>
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <sys/time.h>

#ifndef _PATH_HOSTS
#define	_PATH_HOSTS	"/etc/hosts"
#endif

#ifndef MAXALIASES
#define	MAXALIASES	10
#endif
#ifndef	MAXADDRS
#define	MAXADDRS	20
#endif

#define	WALIGN(x)	(((int)(x) + (sizeof(char *)-1)) & ~(sizeof(char *)-1))
#define	HOSTS_SEP	" \t\n"
#define	ADDRLEN(af)	((af) == AF_INET6 ? sizeof(struct in6_addr) : \
					    sizeof(struct in_addr))
#define	ABUFLEN		sizeof(struct in6_addr)

#ifdef DNS
static struct hostent *dns_ghbyname(const char *name, int af);
static struct hostent *dns_ghbyaddr(const void *addr, int addrlen, int af);
#endif /* DNS */
#ifdef ICMPNL
static struct hostent *icmp_ghbyaddr(const void *addr, int addrlen, int af);
#endif /* ICMPNL */

static struct hostent *
_hpcopy(struct hostent *hp)
{
	static struct hostent *nhp;
	static int nhlen;
	char *cp;
	int len, alen;
	int al_len = 0, hl_len = 0;
	int al_off;
	int i;

	if (hp == NULL) {
		h_errno = HOST_NOT_FOUND;
		return hp;
	}

	/* count length */
	len = sizeof(struct hostent);
	if (hp->h_name)
		len += strlen(hp->h_name) + 1;
	if (hp->h_aliases && hp->h_aliases[0]) {
		for (i = 0; hp->h_aliases[i]; i++)
			len += strlen(hp->h_aliases[i]) + 1;
		al_len = i + 1;
	}
	/* fix alignment */
	len = WALIGN(len);
	al_off = len;
	if (al_len)
		len += sizeof(char *) * al_len;
	alen = hp->h_length;
	if (hp->h_addr_list && hp->h_addr_list[0]) {
		for (i = 0; hp->h_addr_list[i]; i++)
			len += sizeof(char *) + ABUFLEN;
		hl_len = i + 1;
		len += sizeof(char *);	/*NULL*/
	}

	/* copy */
	if (len > nhlen) {
		if (nhp)
			free(nhp);
		nhp = (struct hostent *)malloc(len);
		if (nhp == NULL) {
			h_errno = TRY_AGAIN;
			nhlen = 0;
			return NULL;
		}
		nhlen = len;
	}
	cp = (char *)&nhp[1];
	nhp->h_name = cp;
	strcpy(cp, hp->h_name);
	cp += strlen(cp) + 1;
	if (al_len) {
		nhp->h_aliases = (char **)((char *)nhp + al_off);
		for (i = 0; hp->h_aliases[i]; i++) {
			nhp->h_aliases[i] = cp;
			strcpy(cp, hp->h_aliases[i]);
			cp += strlen(cp) + 1;
		}
		nhp->h_aliases[i] = NULL;
		cp = (char *)&nhp->h_aliases[al_len];
	} else {
		nhp->h_aliases = NULL;
		cp = (char *)nhp + al_off;
	}
	nhp->h_addrtype = hp->h_addrtype;
	nhp->h_length = hp->h_length;
	if (hl_len) {
		nhp->h_addr_list = (char **)cp;
		cp += sizeof(char *) * hl_len;
		for (i = 0; hp->h_addr_list[i]; i++) {
			nhp->h_addr_list[i] = cp;
			memcpy(cp, hp->h_addr_list[i], alen);
			cp += ABUFLEN;
		}
		nhp->h_addr_list[i] = NULL;
	} else
		nhp->h_addr_list = NULL;
	return nhp;
}

static void
_mapaddr(void *addr)
{
	u_char *p = addr;
	int i;

	for (i = 0; i < 4; i++)
		p[12 + i] = p[i];
	for (i = 0; i < 10; i++)
		p[i] = 0;
	p[10] = p[11] = 0xff;
}

static int stayopen;
static FILE *hfp;

static int
_sethtent(int stay)
{
	char *cp;
#ifdef V6ROOT
	char path[1024];
#endif /* V6ROOT */

	if (hfp == NULL) {
#ifdef V6ROOT
		if (cp = getenv("V6ROOT"))
			strcpy(path, cp);
		else
			path[0] = '\0';
		strcat(path, _PATH_HOSTS);
		hfp = fopen(path, "r");
#else /* V6ROOT */
		hfp = fopen(_PATH_HOSTS, "r");
#endif /* V6ROOT */
		if (hfp == NULL)
			return -1;
	} else {
		rewind(hfp);
	}
	stayopen = stay;
	return 0;
}

static int
_endhtent(void)
{
	if (hfp) {
		fclose(hfp);
		hfp = NULL;
	}
	return 0;
}

static struct hostent *
_gethtent(void)
{
	static char buf[BUFSIZ];
	char *p;
	int i;
	static char addrbuf[ABUFLEN];
	static char *alist[MAXALIASES];
	static char *hlist[2];
	static struct hostent hbuf;

	if (hfp == NULL) {
		if (_sethtent(0) < 0)
			return NULL;
	}

	hbuf.h_aliases = alist;
	hbuf.h_addr_list = hlist;
	hlist[0] = addrbuf;
	hlist[1] = NULL;

	while (fgets(buf, sizeof(buf), hfp)) {
		if (buf[0] == '#')
			continue;	/* comment */
		if ((p = strtok(buf, HOSTS_SEP)) == NULL)
			continue;
		if (inet_pton(AF_INET6, p, addrbuf) > 0) {
			hbuf.h_addrtype = AF_INET6;
			hbuf.h_length = sizeof(struct in6_addr);
		} else if (inet_pton(AF_INET, p, addrbuf) > 0) {
			hbuf.h_addrtype = AF_INET;
			hbuf.h_length = sizeof(struct in_addr);
		} else
			continue;
		if ((p = strtok(NULL, HOSTS_SEP)) == NULL)
			continue;	/* no name */
		hbuf.h_name = p;
		for (i = 0; i < MAXALIASES-1; i++)
			if ((alist[i] = strtok(NULL, HOSTS_SEP)) == NULL)
				break;
		if (i == 0)
			hbuf.h_aliases = NULL;
		else
			alist[i] = NULL;
#if 0
printf("_gethtent: %s %s\n", hbuf.h_name, inet_ntop(hbuf.h_addrtype, hbuf.h_addr, 0, 0));
#endif
		return &hbuf;
	}
	if (!stayopen)
		_endhtent();
	return NULL;
}

struct hostent *
gethostbyname2(const char *name, int af)
{
	struct hostent *hp;
	char **cpp;
	char addrbuf[ABUFLEN];

	if (af != AF_INET && af != AF_INET6) {
		h_errno = NO_RECOVERY;
		return NULL;
	}

	/*
	 * TODO: call res_init() to set RES_USE_INET6
	 */

	/*
	 * XXX: This feature is not described in RFC 2133
	 */
	if (inet_pton(af, name, addrbuf) > 0) {
		struct hostent hbuf;
		char *addrs[2];

		if (af == AF_INET && (_res.options & RES_USE_INET6)) {
			af = AF_INET6;
			_mapaddr(addrbuf);
		}
		hp = &hbuf;
		hp->h_name = (char *)name;
		hp->h_aliases = NULL;
		hp->h_addrtype = af;
		hp->h_length = ADDRLEN(af);
		hp->h_addr_list = addrs;
		addrs[0] = addrbuf;
		addrs[1] = NULL;
		h_errno = 0;
		return _hpcopy(hp);
	}

	h_errno = HOST_NOT_FOUND;
	_sethtent(stayopen);
	while (hp = _gethtent()) {
		if (strcmp(hp->h_name, name) != 0) {
			if ((cpp = hp->h_aliases) != NULL) {
				for (; *cpp; cpp++)
					if (strcmp(*cpp, name) == 0)
						goto hp_found;
			}
			continue;
		}
  hp_found:
		if (af && hp->h_addrtype != af) {
			h_errno = NO_DATA;
			continue;
		}
		h_errno = 0;
		break;
	}
	if (!stayopen)
		_endhtent();
#ifdef DNS
	if (!hp)
		hp = dns_ghbyname(name, af);
#endif /* DNS */
	if (hp && af == AF_INET && _res.options & RES_USE_INET6) {
		hp->h_addrtype = AF_INET6;
		hp->h_length = sizeof(struct in6_addr);
		for (cpp = hp->h_addr_list; *cpp; cpp++)
			_mapaddr(*cpp);
	}
	return hp;
}

struct hostent *
gethostbyname(const char *name)
{
	struct hostent *hp = NULL;

	/*
	 * TODO: call res_init() to set RES_USE_INET6
	 */

	if (_res.options & RES_USE_INET6)
		hp = gethostbyname2(name, AF_INET6);
	if (!hp)
		hp = gethostbyname2(name, AF_INET);
	return hp;
}

struct hostent *
gethostbyaddr(const char *src, int len, int af)
{
	struct hostent *hp;
	char **cpp;

	/*
	 * TODO: call res_init() to set RES_USE_INET6
	 */

	if (af != AF_INET && af != AF_INET6) {
		h_errno = NO_RECOVERY;
		return NULL;
	}

	if (af == AF_INET6 && len == sizeof(struct in6_addr)
	&&  (IN6_IS_ADDR_V4MAPPED((struct in6_addr *)src)
	||   IN6_IS_ADDR_V4COMPAT((struct in6_addr *)src))) {
		src += sizeof(struct in6_addr) - sizeof(struct in_addr);
		af = AF_INET;
		len = sizeof(struct in_addr);
	}
	h_errno = HOST_NOT_FOUND;
	_sethtent(stayopen);
	while (hp = _gethtent()) {
		if (hp->h_addrtype != af)
			continue;
		if (hp->h_length != len)
			continue;
		if (!bcmp(hp->h_addr_list[0], src, len)) {
			h_errno = 0;
			break;
		}
	}
	if (!stayopen)
		_endhtent();
#ifdef ICMPNL
	if (!hp)
		hp = icmp_ghbyaddr(src, len, af);
#endif /* ICMPNL */
#ifdef DNS
	if (!hp)
		hp = dns_ghbyaddr(src, len, af);
#endif /* DNS */
	if (hp && af == AF_INET && _res.options & RES_USE_INET6) {
		hp->h_addrtype = AF_INET6;
		hp->h_length = sizeof(struct in6_addr);
		for (cpp = hp->h_addr_list; *cpp; cpp++)
			_mapaddr(*cpp);
	}
	return hp;
}

#ifdef DNS

#include <arpa/nameser.h>
#include <resolv.h>

#ifdef DEBUG
#define	ASSERT(X)	if (!(X)) fprintf(stderr, "ASSFAIL: %s %d: %s\n", __FILE__, __LINE__, #X);
#else
#define	ASSERT(X)
#endif

static struct hostent *
dns_ghbyname(const char *name, int af)
{
	int n;
	u_char answer[BUFSIZ];
	char tbuf[MAXDNAME+1];
	HEADER *hp;
	u_char *cp, *eom;
	int qtype;
	int type, class, ancount, qdcount;
	u_long ttl;
	char hostbuf[BUFSIZ];
	char *bp;
	char *alist[MAXALIASES];
	char *hlist[MAXADDRS];
	struct hostent hbuf;
	int buflen;
	int na, nh;

	if ((_res.options & RES_INIT) == 0) {
		if (res_init() < 0)
			return NULL;
	}
#ifdef DEBUG
	_res.options |= RES_DEBUG;
	printf(";; res retry %d\n", _res.retry);
	printf(";; res nscount %d\n", _res.nscount);
#endif
	hbuf.h_aliases = NULL;
	hbuf.h_addrtype = af;
	hbuf.h_length = ADDRLEN(af);
	hbuf.h_addr_list = hlist;
	na = nh = 0;

	qtype = (af == AF_INET6 ? T_AAAA : T_A);
	n = res_search(name, C_IN, qtype, answer, sizeof(answer));
	if (n < 0)
		return NULL;
	hp = (HEADER *)answer;
	eom = answer + n;
	ancount = ntohs(hp->ancount);
	qdcount = ntohs(hp->qdcount);
	ASSERT(qdcount == 1);
	cp = answer + sizeof(HEADER);
	bp = hostbuf;
	buflen = sizeof(hostbuf);

	n = dn_expand(answer, eom, cp, bp, buflen);
	ASSERT(n >= 0);
	cp += n + QFIXEDSZ;
	hbuf.h_name = bp;
	n = strlen(bp) + 1;
	bp += n;
	buflen -= n;
	while (ancount-- > 0 && cp < eom) {
		n = dn_expand(answer, eom, cp, bp, buflen);
		ASSERT(n >= 0);
		cp += n;		/* name */
		type = _getshort(cp);
		cp += 2;		/* type */
		class = _getshort(cp);
		cp += 2;		/* class */
		ttl = _getlong(cp);
		cp += 4;		/* ttl */
		n = _getshort(cp);
		cp += 2;		/* len */
		ASSERT(class == C_IN);
		switch (type) {
		case T_CNAME:
			if (na >= MAXALIASES-1) {
				cp += n;
				break;
			}
			n = dn_expand(answer, eom, cp, tbuf, sizeof(tbuf));
			ASSERT(n >= 0);
			cp += n;
			/* alias */
			if (na == 0)
				hbuf.h_aliases = alist;
			alist[na++] = bp;
			n = strlen(bp) + 1;
			bp += n;
			buflen -= n;
			/* canon */
			n = strlen(tbuf) + 1;
			ASSERT(n < buflen);
			strcpy(bp, tbuf);
			hbuf.h_name = bp;
			bp += n;
			buflen -= n;
			break;
		case T_A:
		case T_AAAA:
if (type != qtype) { printf("Got other type!\n"); cp += n; break; }
			bp = (char *)WALIGN(bp);
			ASSERT(n == hbuf.h_length);
			ASSERT(n < buflen);
			if (nh < MAXADDRS-1) {
				hlist[nh++] = bp;
				memcpy(bp, cp, n);
				bp += n;
				buflen -= n;
			}
			cp += n;
			break;
		default:
			ASSERT(0);
			cp += n;
			break;
		}
	}
	if (nh == 0) {
		h_errno = NO_RECOVERY;
		/* XXX: should be test if NO_DATA */
		return NULL;
	}
	alist[na] = NULL;
	hlist[nh] = NULL;
	return _hpcopy(&hbuf);
}

static struct hostent *
dns_ghbyaddr(const void *addr, int addrlen, int af)
{
	int n;
	u_char answer[BUFSIZ];
	HEADER *hp;
	u_char c, *cp, *eom;
	int type, class, ancount, qdcount;
	u_long ttl;
	char hostbuf[BUFSIZ];
	char *bp;
	char *alist[MAXALIASES];
	char *hlist[2];
	struct hostent hbuf;
	int buflen;
	int na;
	static char hex[] = "0123456789abcdef";

	/* XXX */
	if (af == AF_INET6 && IN6_IS_ADDR_LINKLOCAL((struct in6_addr *)addr))
		return NULL;

	if ((_res.options & RES_INIT) == 0) {
		if (res_init() < 0)
			return NULL;
	}
	hbuf.h_name = NULL;
	hbuf.h_aliases = NULL;
	hbuf.h_addrtype = af;
	hbuf.h_length = addrlen;
	hbuf.h_addr_list = hlist;
	hlist[0] = (char *)addr;
	hlist[1] = NULL;
	na = 0;

	if (af == AF_INET6) {
		for (n = 0, cp = (u_char *)addr+sizeof(struct in6_addr)-1,
					bp = hostbuf;
				n < sizeof(struct in6_addr);
				n++, cp--, bp += 4) {
			bp[0] = hex[*cp & 0xf];
			bp[1] = '.';
			bp[2] = hex[*cp >> 4];
			bp[3] = '.';
		}
		strcpy(bp, "ip6.int");
	} else {
		for (n = 0, cp = (u_char *)addr+sizeof(struct in_addr)-1,
					bp = hostbuf;
				n < sizeof(struct in_addr);
				n++, cp--) {
			c = *cp;
			if (c >= 100) {
				*bp++ = '0' + c / 100;
				c %= 100;
			}
			if (c >= 10) {
				*bp++ = '0' + c / 10;
				c %= 10;
			} else if (*cp >= 100) {
				*bp++ = '0';
			}
			*bp++ = '0' + c;
			*bp++ = '.';
		}
		strcpy(bp, "in-addr.arpa");
	}

	n = res_search(hostbuf, C_IN, T_PTR, answer, sizeof(answer));
	if (n < 0) {
		return NULL;
	}
	hp = (HEADER *)answer;
	eom = answer + n;
	ancount = ntohs(hp->ancount);
	qdcount = ntohs(hp->qdcount);
	ASSERT(qdcount == 1);
	cp = answer + sizeof(HEADER);
	bp = hostbuf;
	buflen = sizeof(hostbuf);

	n = dn_expand(answer, eom, cp, bp, buflen);
	ASSERT(n >= 0);
	cp += n + QFIXEDSZ;
	while (ancount-- > 0 && cp < eom) {
		n = dn_expand(answer, eom, cp, bp, buflen);
		ASSERT(n >= 0);
		cp += n;		/* name */
		type = _getshort(cp);
		cp += 2;		/* type */
		class = _getshort(cp);
		cp += 2;		/* class */
		ttl = _getlong(cp);
		cp += 4;		/* ttl */
		n = _getshort(cp);
		cp += 2;		/* len */
		ASSERT(class == C_IN);
		switch (type) {
		case T_PTR:
			n = dn_expand(answer, eom, cp, bp, buflen);
			ASSERT(n >= 0);
			cp += n;
			if (na >= MAXALIASES-1)
				break;
			if (hbuf.h_name == NULL) {
				hbuf.h_name = bp;
			} else {
				if (na == 0)
					hbuf.h_aliases = alist;
				alist[na++] = bp;
			}
			n = strlen(bp) + 1;
			bp += n;
			buflen -= n;
			break;
		default:
			ASSERT(0);
			/*FALLTHRU*/
		case T_CNAME:
			cp += n;
			break;
		}
	}
	if (hbuf.h_name == NULL)
		return NULL;
	alist[na] = NULL;
	return _hpcopy(&hbuf);
}
#endif /* DNS */

#ifdef ICMPNL
#include <sys/uio.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>

static char *
icmp_fqdn_query(const struct in6_addr *addr, int ifindex)
{
	int s;
	struct icmp6_filter filter;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct in6_pktinfo *pkt;
	char cbuf[256];
	char buf[1024];
	int cc;
	struct icmp6_fqdn_query *fq;
	struct icmp6_fqdn_reply *fr;
	static struct host_cache {
		struct host_cache *hc_next;
		int hc_ifindex;
		struct in6_addr hc_addr;
		char *hc_name;
	} *hc_head = NULL;
	struct host_cache *hc;
	static int pid;
	struct sockaddr_in6 sin6;
	struct iovec iov;
	fd_set s_fds, fds;
	struct timeval tout;
	int len;
	char *name;

	for (hc = hc_head; hc; hc = hc->hc_next) {
		if (hc->hc_ifindex == ifindex
		&&  IN6_ARE_ADDR_EQUAL(&hc->hc_addr, addr))
			return hc->hc_name;
	}

	hc = (struct host_cache *)malloc(sizeof(*hc));
	if (hc == NULL)
		return NULL;
	hc->hc_ifindex = ifindex;
	hc->hc_addr = *addr;
	hc->hc_name = NULL;
	hc->hc_next = hc_head;
	hc_head = hc;

	if (pid == 0)
		pid = getpid();

	ICMP6_FILTER_SETBLOCKALL(&filter);
	ICMP6_FILTER_SETPASS(ICMP6_FQDN_REPLY, &filter);

	FD_ZERO(&s_fds);
	tout.tv_sec = 0;
	tout.tv_usec = 200000;	/*XXX: 200ms*/

	fq = (struct icmp6_fqdn_query *)buf;
	fq->icmp6_fqdn_type = ICMP6_FQDN_QUERY;
	fq->icmp6_fqdn_code = 0;
	fq->icmp6_fqdn_cksum = 0;
	fq->icmp6_fqdn_id = (u_short)pid;
	fq->icmp6_fqdn_unused = 0;
	fq->icmp6_fqdn_cookie[0] = 0;
	fq->icmp6_fqdn_cookie[1] = 0;

	memset(&sin6, 0, sizeof(sin6));
	sin6.sin6_family = AF_INET6;
	sin6.sin6_addr = *addr;

	memset(&msg, 0, sizeof(msg));
	msg.msg_name = (caddr_t)&sin6;
	msg.msg_namelen = sizeof(sin6);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	iov.iov_base = (caddr_t)buf;
	iov.iov_len = sizeof(struct icmp6_fqdn_query);

	if (ifindex) {
		msg.msg_control = cbuf;
		msg.msg_controllen = sizeof(cbuf);
		cmsg = CMSG_FIRSTHDR(&msg);
		cmsg->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
		cmsg->cmsg_level = IPPROTO_IPV6;
		cmsg->cmsg_type = IPV6_PKTINFO;
		pkt = (struct in6_pktinfo *)&cmsg[1];
		memset(&pkt->ipi6_addr, 0, sizeof(struct in6_addr));
		pkt->ipi6_ifindex = ifindex;
		cmsg = CMSG_NXTHDR(&msg, cmsg);
		msg.msg_controllen = (char *)cmsg - cbuf;
	}

	s = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
	if (s < 0)
		return NULL;
	(void)setsockopt(s, IPPROTO_ICMPV6, ICMP6_FILTER,
			 (char *)&filter, sizeof(filter));
	cc = sendmsg(s, &msg, 0);
	if (cc < 0) {
		close(s);
		return NULL;
	}
	FD_SET(s, &s_fds);
	for (;;) {
		fds = s_fds;
		if (select(s + 1, &fds, NULL, NULL, &tout) <= 0) {
			close(s);
			return NULL;
		}
		len = sizeof(sin6);
		cc = recvfrom(s, buf, sizeof(buf), 0,
			      (struct sockaddr *)&sin6, &len);
		if (cc <= 0) {
			close(s);
			return NULL;
		}
		if (cc < sizeof(struct ip6_hdr) + sizeof(struct icmp6_hdr))
			continue;
		if (!IN6_ARE_ADDR_EQUAL(addr, &sin6.sin6_addr))
			continue;
		fr = (struct icmp6_fqdn_reply *)(buf + sizeof(struct ip6_hdr));
		if (fr->icmp6_fqdn_type == ICMP6_FQDN_REPLY)
			break;
	}
	close(s);
	if (fr->icmp6_fqdn_cookie[1] != 0) {
		/* rfc1788 type */
		name = buf + sizeof(struct ip6_hdr) + sizeof(struct icmp6_hdr) + 4;
		len = (buf + cc) - name;
	} else {
		len = fr->icmp6_fqdn_namelen;
		name = fr->icmp6_fqdn_name;
	}
	if (len <= 0)
		return NULL;
	name[len] = 0;

	hc->hc_name = strdup(name);
	return hc->hc_name;
}

static struct hostent *
icmp_ghbyaddr(const void *addr, int addrlen, int af)
{
	struct hostent hbuf, *hp;
	char *hname, *alist[2];
	int ifindex;
	struct in6_addr addr6;

	if (af != AF_INET6) {
		/*
		 * Note: rfc1788 defines Who Are You for IPv4,
		 * but no one implements it.
		 */
		return NULL;
	}

	memcpy(&addr6, addr, addrlen);
	ifindex = addr6.s6_addr[3];
	addr6.s6_addr[3] = 0;

	if (!IN6_IS_ADDR_LINKLOCAL(&addr6))
		return NULL;	/*XXX*/

	hname = icmp_fqdn_query(&addr6, ifindex);
	if (hname == NULL)
		return NULL;

	hp = &hbuf;
	hp->h_name = hname;
	hp->h_aliases = NULL;
	hp->h_addrtype = af;
	hp->h_length = addrlen;
	hp->h_addr_list = alist;
	alist[0] = (char *)addr;
	alist[1] = NULL;
	return _hpcopy(hp);
}
#endif /* ICMPNL */
