/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <kvm.h>
#include <nlist.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet6/in6_var.h>
#include <netinet6/ip6.h>
#include <netinet6/ip6_var.h>
#include <netinet6/in6_pcb.h>
#include <netinet6/udp6.h>
#include <netinet6/udp6_var.h>

#include <netinet/in.h>
#include <arpa/inet.h>

kvm_t	*kvmd;

struct	nlist nl[] = {
#define	N_UDB6	0
	{ "_udb6" },
	"",
};

char *ifname __P((struct ifnet *));
char *in6flag __P((int));
void mc __P((struct in6_multi *));
void kread __P((u_long, void *, int));

#define	KREAD(addr, buf, type) \
	kread((u_long)addr, (void *)buf, sizeof(type))

const char *inet6_n2a(p)
	struct in6_addr *p;
{
	static char buf[BUFSIZ];

	if (IN6_IS_ADDR_ANY(p))
		return "*";
	return inet_ntop(AF_INET6, (void *)p, buf, sizeof(buf));
}

void main()
{
	char	buf[BUFSIZ];
	struct	in6pcb	udb6;
	struct	ip6_moptions ip6m;

	if ((kvmd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, buf)) == NULL) {
		perror("kvm_openfiles");
		exit(1);
	}
	if (kvm_nlist(kvmd, nl) < 0) {
		perror("kvm_nlist");
		exit(1);
	}
	if (nl[N_UDB6].n_value == 0) {
		printf("symbol _udb6 not found\n");
		exit(1);
	}
	KREAD(nl[N_UDB6].n_value, &udb6, struct in6pcb);
	while (udb6.in6p_next &&
		udb6.in6p_next != (struct in6pcb *)nl[N_UDB6].n_value) {
		KREAD(udb6.in6p_next, &udb6, struct in6pcb);
		printf("local %s.%d", inet6_n2a(&udb6.in6p_laddr),
			ntohs(udb6.in6p_lport) & 0xffff);
		printf(" foreign %s.%d", inet6_n2a(&udb6.in6p_faddr),
			ntohs(udb6.in6p_fport) & 0xffff);
		printf(" %s", in6flag(udb6.in6p_flags));
		printf(" opt(%x)", udb6.in6p_options);
		printf("\n");
		if (udb6.in6p_moptions) {
			int i;

			KREAD(udb6.in6p_moptions, &ip6m, struct ip6_moptions);
			printf("\t");
			if (ip6m.im6o_multicast_ifp)
				printf("ifname %s ",
					ifname(ip6m.im6o_multicast_ifp));
			printf("mhlim(%d) ", ip6m.im6o_multicast_hlim);
			printf("mloop(%d) ", ip6m.im6o_multicast_loop);
			printf("nm(%d)", ip6m.im6o_num_memberships);
			printf("\n");
			for (i = 0; i < ip6m.im6o_num_memberships; i++) {
				mc(ip6m.im6o_membership[i]);
			}
		}
	}
}

void mc(imp)
	struct in6_multi *imp;
{
	struct in6_multi im;

	KREAD(imp, &im, struct in6_multi);
	printf("\t\t%s", ifname(im.in6m_ifp));
	printf(" %s", inet6_n2a(&im.in6m_addr));
	printf(" ref(%d)", im.in6m_refcount);
	printf(" timer(%d)", im.in6m_timer);
	printf("\n");
}

char *ifname(ifp)
	struct ifnet *ifp;
{
	struct ifnet ifnet;
	char ifbuf[BUFSIZ];
	static char buf[BUFSIZ];

	KREAD(ifp, &ifnet, struct ifnet);
	KREAD(ifnet.if_name, ifbuf, ifbuf);
	sprintf(buf, "%s%d", ifbuf, ifnet.if_unit);
	return buf;
}

void kread(addr, buf, len)
	u_long addr;
	void *buf;
	int len;
{
	if (kvm_read(kvmd, addr, buf, len) != len) {
		perror("kvm_read");
		exit(1);
	}
}

char *in6flag(flags)
	int flags;
{
	static char buf[BUFSIZ];

	strcpy(buf, "");
	if (flags == 0)
		return "";
#define	FLAG(x, y) if (flags & x) strcat(buf, y)
	FLAG(IN6P_RECVOPTS, " Opts");
	FLAG(IN6P_RECVRETOPTS, " Opts");
	FLAG(IN6P_RECVDSTADDR, " DstAddr");
/*
	FLAG(IN6P_HDRINCL, " HdrIncl");
*/
	return buf + 1;
}
