/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/socket.h>
#ifdef ADVAPI
#include <sys/uio.h>
#endif

#include <net/if.h>
#include <netinet/in.h>
#include <netinet6/ip6.h>
#include <netinet6/ip6_var.h>
#include <netinet6/icmp6.h>
#include <unistd.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include "advcap.h"

#define ALLNODES "ff02::1"
#define ALLROUTERS "ff02::2"
#define ANY "::"
#define RTSOLLEN 48

struct	icmp_ra {
	u_char	icmp_type;
	u_char	icmp_code;
	u_short	icmp_cksum;
	u_char	icmp_chlim;
	u_char	icmp_raflags;
	u_short	icmp_rltime;
	u_long	icmp_rtime;
	u_long	icmp_retrans;
};

struct	rainfo {
	struct	rainfo *next;
	u_short if_index;
	struct  sockaddr_in6 saddr;
	struct	icmp_ra icmp;
	struct	nd_opt_prefix_info pinfo;
	struct	nd_opt_mtu mtu;
} *ralist = NULL;

int sock;

int debug = 0;

void getconfig __P((char *));
void sock_open __P((void));
void reply __P((void));
long rgetnum __P((char *));

int
main(argc, argv)
	int argc;
	char *argv[];
{
	fd_set fdset;
	struct timeval  timeout;
	int i;
	
	sock_open();
	--argc;
	++argv;
	if (argc == 0) {
		fprintf(stderr, "usage: rtadvd interfaces...\n");
		exit(1);
	}
	while (argc--)
		getconfig(*argv++);
	all_output();

	if (!debug)
		daemon(1,0);

	timeout.tv_sec = 600;
	timeout.tv_usec = 0;
	FD_ZERO(&fdset);
	while (1) {
		FD_SET(sock, &fdset);
		if ((i = select(sock + 1, &fdset, NULL, NULL, &timeout)) < 0)
			perror("rtadvd: select");
		if (i == 0)
			all_output();
		else
			reply();
	}
	exit(0);
}

void
reply()
{
	int i;
	struct sockaddr_in6 from;
	int from_len;
	u_char answer[576];
	struct ip6_hdr *ip;
	struct icmp6_hdr *icp;
	struct rainfo *ra = ralist;
	int ifindex;
#ifdef ADVAPI
	struct msghdr m;
	struct cmsghdr *cm;
	struct iovec iov[2];
	u_char cmsgbuf[256];
	struct in6_pktinfo *pi;
#endif /*ADVAPI*/

	from_len = sizeof(from);
#ifdef ADVAPI
	m.msg_name = (caddr_t)&from;
	m.msg_namelen = from_len;
	iov[0].iov_base = (caddr_t)answer;
	iov[0].iov_len = sizeof(answer);
	m.msg_iov = iov;
	m.msg_iovlen = 1;
	cm = (struct cmsghdr *)cmsgbuf;
	m.msg_control = (caddr_t) cm;
	m.msg_controllen = sizeof(cmsgbuf);
	if ((i = recvmsg(sock, &m, 0)) < 0)
		return;
	from_len = m.msg_namelen;
	ifindex = 0;
	for (cm = (struct cmsghdr *)CMSG_FIRSTHDR(&m);
	     cm;
	     cm = (struct cmsghdr *)CMSG_NXTHDR(&m, cm)) {
		if (cm->cmsg_level == IPPROTO_IPV6
		 && cm->cmsg_type == IPV6_PKTINFO) {
			pi = (struct in6_pktinfo *)(CMSG_DATA(cm));
			ifindex = pi->ipi6_ifindex;
			break;
		}
	}
#else
	if ((i = recvfrom(sock, answer, sizeof(answer), 0,
			  (struct sockaddr *)&from,
			  &from_len)) < 0)
		return;
	if (IN6_IS_ADDR_LINKLOCAL(&from->sin6_addr))
		ifindex = ntohs(from.sin6_addr.s6_addr16[1]);
	else
		ifindex = 0;
#endif /*ADVAPI*/
	if (i < RTSOLLEN)
		return;
	ip = (struct ip6_hdr *)answer;
	icp = (struct icmp6_hdr *)(ip + 1);		/* XXX header chain */

	/*
	 * verification - discovery-v2-00 6.1.1
	 */
	if (ip->ip6_hlim != 255)
		return;
	if (icp->icmp6_type != ND_ROUTER_SOLICIT)
		return;
	/* code is checked in kernel */
	/* len is checked in kernel */
	/* XXX verification of options */
	if (IN6_IS_ADDR_UNSPECIFIED(&ip->ip6_src)) {
		/* XXX if lladdr option is not there, ignore the packet */
	}

	/* random sleep() should be added; discovery-v2-00 6.2.6 */

	while (ra != NULL) {
		if (ifindex == ra->if_index)
			ra_output(ra);
		ra = ra->next;
	}
}

void
getconfig(intface)
	char *intface;
{
	int stat;
	char tbuf[BUFSIZ];
	struct rainfo *tmp;
	struct ipv6_mreq mreq;
	u_char flag;
	char buf[BUFSIZ];
	char *bp = buf;
	char *addr;

#define MUSTHAVE(var, cap)	\
    {									\
	int t;								\
	if ((t = agetnum(cap)) < 0) {					\
		fprintf(stderr, "rtadvd: need %s for interface %s\n",	\
			cap, intface);					\
		exit(1);						\
	}								\
	var = t;							\
     }
#define MAYHAVE(var, cap, def)	\
     {									\
	if ((var = agetnum(cap)) < 0)					\
		var = def;						\
     }

	if ((stat = agetent(tbuf, intface)) <= 0) {
		fprintf(stderr, "rtadvd: don't know %s\n", intface);
		exit(1);
	}

	tmp = (struct rainfo *)malloc(sizeof(*ralist));
	tmp->if_index = (u_short)if_nametoindex(intface);
	if (tmp->if_index == 0) {
		fprintf(stderr, "rtadvd: unknown interface %s\n", intface);
		exit(1);
	}

	memset(&tmp->saddr, 0, sizeof(struct sockaddr_in6));
	if (inet_pton(AF_INET6, ALLNODES, &tmp->saddr.sin6_addr.s6_addr) != 1) {
		perror("inet_pton");
		exit(1);
	}
	tmp->saddr.sin6_len = sizeof(struct sockaddr_in6);
	tmp->saddr.sin6_family = AF_INET6;
#ifndef ADVAPI
	tmp->saddr.sin6_addr.u6_addr.u6_addr16[1] = htons(tmp->if_index);
#endif

	tmp->icmp.icmp_type = ND_ROUTER_ADVERT;
	tmp->icmp.icmp_code = 0;
	MUSTHAVE(tmp->icmp.icmp_chlim, "chlim");
	MUSTHAVE(tmp->icmp.icmp_raflags, "raflags");
	tmp->icmp.icmp_raflags &= 0xc0;
	MUSTHAVE(tmp->icmp.icmp_rltime, "rltime");
	tmp->icmp.icmp_rltime = htons(tmp->icmp.icmp_rltime);
	MUSTHAVE(tmp->icmp.icmp_rtime, "rtime");
	tmp->icmp.icmp_rtime = htonl(tmp->icmp.icmp_rtime);
	MUSTHAVE(tmp->icmp.icmp_retrans, "retrans");
	tmp->icmp.icmp_retrans = htonl(tmp->icmp.icmp_retrans);
	tmp->pinfo.nd_opt_pi_type = ND_OPT_PREFIX_INFORMATION;
	tmp->pinfo.nd_opt_pi_len =
		(sizeof(struct nd_opt_prefix_info) >> 3) & 0xff;
	MUSTHAVE(tmp->pinfo.nd_opt_pi_prefix_len, "prefixlen");
	MUSTHAVE(tmp->pinfo.nd_opt_pi_flags_reserved, "pinfoflags");
	MUSTHAVE(tmp->pinfo.nd_opt_pi_valid_time, "vltime");
	tmp->pinfo.nd_opt_pi_valid_time =
		htonl(tmp->pinfo.nd_opt_pi_valid_time );
	MUSTHAVE(tmp->pinfo.nd_opt_pi_preferred_time, "pltime");
	tmp->pinfo.nd_opt_pi_preferred_time =
		htonl(tmp->pinfo.nd_opt_pi_preferred_time);
	tmp->pinfo.nd_opt_pi_reserved2 = 0;
	addr = agetstr("addr", &bp);
	if (addr == NULL) {
		fprintf(stderr, "rtadvd: need %s for interface %s\n",
			"addr", intface);
		exit(1);
	}
	if (inet_pton(AF_INET6, addr, &tmp->pinfo.nd_opt_pi_prefix) != 1) {
		perror("inet_pton");
		exit(1);
	}
	tmp->mtu.nd_opt_mtu_type = 5;
	tmp->mtu.nd_opt_mtu_len = 1;
	tmp->mtu.nd_opt_mtu_reserved= 0;     
	MUSTHAVE(tmp->mtu.nd_opt_mtu_mtu, "mtu");
	tmp->mtu.nd_opt_mtu_mtu = htonl(tmp->mtu.nd_opt_mtu_mtu);

	if (inet_pton(AF_INET6, ALLROUTERS, &mreq.ipv6mr_multiaddr.s6_addr)
			!= 1)
		exit(1);
#ifndef ADVAPI
	mreq.ipv6mr_multiaddr.u6_addr.u6_addr16[1] = htons(tmp->if_index);
#endif
	mreq.ipv6mr_interface = tmp->if_index;
	if (setsockopt(sock, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &mreq,
			sizeof(mreq)) < 0) {
		perror("IPV6_ADD_MEMBERSHIP");
	}

	/* okey */
	tmp->next = ralist;
	ralist = tmp;
}

void
sock_open()
{
	struct sockaddr_in6 me;
	u_int hlim = 255;
	struct icmp6_filter filt;
	
	if ((sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0)
		err(1, "socket");

	memset(&me, 0, sizeof(struct sockaddr_in6));
	me.sin6_len = sizeof(struct sockaddr_in6);
	me.sin6_family = AF_INET6;

#if 0
	if (bind(sock, (struct sockaddr *)&me, sizeof(me)) < 0)
		err(1, "bind");
#endif

	if (setsockopt(sock, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &hlim,
		      sizeof(hlim)) == -1)	
		err(1, "IP6_MULTICAST_HOPS");

#ifdef ICMP6_FILTER
	ICMP6_FILTER_SETBLOCKALL(&filt);
	ICMP6_FILTER_SETPASS(ND_ROUTER_SOLICIT, &filt);
	if (setsockopt(sock, IPPROTO_ICMPV6, ICMP6_FILTER, &filt,
			sizeof(filt)) < 0)
		err(1, "ICMP6_FILTER");
#endif /*ICMP6_FILTER*/

#ifdef ADVAPI
    {
	int on = 1;
	if (setsockopt(sock, IPPROTO_IPV6, IPV6_PKTINFO, &on, sizeof(on)) < 0)
		err(1, "IPV6_PKTINFO");
    }
#endif /*ADVAPI*/
}

int
all_output()
{
	struct rainfo *ra = ralist;

	while (ra != NULL) {
		ra_output(ra);
		ra = ra->next;
	}
}

int
ra_output(rainfo)
struct rainfo *rainfo;
{
	struct sockaddr_in6 to;
	struct icmp6_hdr *icp;
	u_int hlim = 255;
	u_char outpack[sizeof(struct nd_router_advert)];
	u_short index;
	int i, cc = sizeof(struct nd_router_advert)
		+ sizeof(struct nd_opt_prefix_info)
		+ sizeof(struct nd_opt_mtu);
#ifdef ADVAPI
	struct msghdr m;
	struct cmsghdr *cm;
	struct iovec iov[2];
	u_char cmsgbuf[256];
	struct in6_pktinfo *pi;

	m.msg_name = (caddr_t)&rainfo->saddr;
	m.msg_namelen = sizeof(rainfo->saddr);
	iov[0].iov_base = (caddr_t)&rainfo->icmp;
	iov[0].iov_len = cc;
	m.msg_iov = iov;
	m.msg_iovlen = 1;
	if (!rainfo->if_index) {
		m.msg_control = NULL;
		m.msg_controllen = 0;
	} else {
		memset(cmsgbuf, 0, sizeof(cmsgbuf));
		cm = (struct cmsghdr *)cmsgbuf;
		m.msg_control = (caddr_t)cm;
		m.msg_controllen = CMSG_SPACE(sizeof(struct in6_pktinfo));

		cm->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
		cm->cmsg_level = IPPROTO_IPV6;
		cm->cmsg_type = IPV6_PKTINFO;
		pi = (struct in6_pktinfo *)CMSG_DATA(cm);
		memset(&pi->ipi6_addr, 0, sizeof(pi->ipi6_addr));	/*XXX*/
		pi->ipi6_ifindex = rainfo->if_index;
	}
	i = sendmsg(sock, &m, 0);
#else
	i = sendto(sock, (char *)&rainfo->icmp, cc, 0,
		   (struct sockaddr *)&rainfo->saddr, sizeof(rainfo->saddr));
#endif /*ADVAPI*/

	if (i < 0 || i != cc)  {
		if (i < 0)
			perror("rtadvd: ra_output");
	}
}
